Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq rels ordtype.
Require Import prelude domains heaps.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(* partial commutative monoid *)

Module PCM.

Record mixin_of (T : Type) := Mixin {
    valid_op : T -> bool;
    join_op : T -> T -> T;
    unit_op : T;
    _ : commutative join_op;
    _ : associative join_op;
    _ : left_id unit_op join_op;
    (* cancellative *)
(*    _ : forall x y z, 
          valid_op (join_op x y) -> join_op x y = join_op x z -> y = z; *)
    (* monotonicity of valid *)
    _ : forall x y, valid_op (join_op x y) -> valid_op x}.

Section ClassDef.

Record class_of T := Class {mixin : mixin_of T}.

Structure type : Type := Pack {sort : Type; _ : class_of sort; _ : Type}.
Local Coercion sort : type >-> Sortclass.

Variables (T : Type) (cT : type).
Definition class := let: Pack _ c _ as cT' := cT return class_of cT' in c.
Definition clone c of phant_id class c := @Pack T c T.

(* produce a PCM type out of the mixin *)
(* equalize m0 and m by means of a phantom *)
Definition pack (m0 : mixin_of T) := 
  fun m & phant_id m0 m => Pack (@Class T m) T.

Definition valid := valid_op (mixin class).
Definition join := join_op (mixin class).
Definition Unit := unit_op (mixin class).

End ClassDef.

Module Exports.
Coercion sort : type >-> Sortclass.
Notation pcm := PCM.type.
Notation PCMMixin := PCM.Mixin.
Notation PCM T m := (@pack T _ m id).

Notation "[ 'pcm' 'of' T 'for' cT ]" := (@clone T cT _ id)
  (at level 0, format "[ 'pcm'  'of'  T  'for'  cT ]") : form_scope.
Notation "[ 'pcm' 'of' T ]" := (@clone T _ _ id)
  (at level 0, format "[ 'pcm'  'of'  T ]") : form_scope.

Notation "x \+ y" := (join x y) (at level 43, left associativity).
Notation valid := PCM.valid.
Notation Unit := PCM.Unit.

Implicit Arguments PCM.Unit [cT].

(* Restating the laws, with the notation. *) 
(* Plus some additional derived lemmas.   *)

Section Laws.
Variable U : pcm.

Lemma joinC (x y : U) : x \+ y = y \+ x.
Proof. by case: U x y=>tp [[v j z Cj *]]; apply: Cj. Qed.

Lemma joinA (x y z : U) : x \+ (y \+ z) = x \+ y \+ z.
Proof. by case: U x y z=>tp [[v j z Cj Aj *]]; apply: Aj. Qed.

Lemma joinAC (x y z : U) : x \+ y \+ z = x \+ z \+ y.
Proof. by rewrite -joinA (joinC y) joinA. Qed.

Lemma joinCA (x y z : U) : x \+ (y \+ z) = y \+ (x \+ z).
Proof. by rewrite joinA (joinC x) -joinA. Qed.

(*
Lemma joinKr (x y z : U) : valid (x \+ y) -> x \+ y = x \+ z -> y = z.
Proof. by case: U x y z=>tp [[v j z Cj Aj Uj /= Kj Mj inv f]]; apply: Kj. Qed.

Lemma joinKl (x y z : U) : valid (x \+ z) -> x \+ z = y \+ z -> x = y.
Proof. by rewrite -!(joinC z); apply: joinKr. Qed.
*)

Lemma validL (x y : U) : valid (x \+ y) -> valid x.
Proof. case: U x y=>tp [[v j z Cj Aj Uj /= Mj inv f]]; apply: Mj. Qed.

Lemma validR (x y : U) : valid (x \+ y) -> valid y.
Proof. by rewrite joinC; apply: validL. Qed.

Lemma unitL (x : U) : Unit \+ x = x.
Proof. by case: U x=>tp [[v j z Cj Aj Uj *]]; apply: Uj. Qed.

Lemma unitR (x : U) : x \+ Unit = x.
Proof. by rewrite joinC unitL. Qed.

End Laws.

End Exports.

End PCM.

Export PCM.Exports.

(* product resource algebras *)
(* will be useful for combining all the ghosts belonging *)
(* to one and the same process *)

Section PairPCM.
Variables (R1 R2 : pcm).
Local Notation tp := (R1 * R2)%type.

Definition pvalid := [fun x : tp => valid x.1 && valid x.2].
Definition pjoin := [fun x1 x2 : tp => (x1.1 \+ x2.1, x1.2 \+ x2.2)].
Definition punit : tp := (Unit, Unit).

Local Notation "x \+ y" := (pjoin x y) (at level 43, left associativity).

Lemma pjoinC (x y : tp) : x \+ y = y \+ x.
Proof. 
move: x y => [x1 x2][y1 y2].
by rewrite /pjoin /= (joinC x1) (joinC x2). 
Qed.

Lemma pjoinA (x y z : tp) : x \+ (y \+ z) = x \+ y \+ z.
Proof.
move: x y z => [x1 x2][y1 y2][z1 z2] /=.
by rewrite /pjoin (joinA x1) (joinA x2).
Qed.

(*
Lemma pjoinKr (x y z : tp) : 
        pvalid (x \+ y) -> x \+ y = x \+ z -> y = z.
Proof.
rewrite /pvalid /pjoin; move: x y z => [x1 x2][y1 y2][z1 z2] /=.
case/andP=>D1 D2; case.
by move/(joinKr D1)=>->; move/(joinKr D2)=>->.
Qed.
*)

Lemma pvalidL (x y : tp) : pvalid (x \+ y) -> pvalid x.
Proof.
move: x y => [x1 x2][y1 y2] /=.
case/andP=>D1 D2.
by rewrite /pvalid (validL D1) (validL D2).
Qed.

Lemma punitL (x : tp) : punit \+ x = x.
Proof. by case: x=>x1 x2; rewrite /= !unitL. Qed.

Definition pairPCMMixin := PCMMixin pjoinC pjoinA punitL pvalidL.
Canonical pairPCM := Eval hnf in PCM (R1 * R2) pairPCMMixin.

End PairPCM.


