Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq rels ordtype.
Require Import Eqdep prelude heaps pcm resources8.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.
Set Automatic Coercions Import.

(***********)
(* Actions *)
(***********)

Delimit Scope proc_scope with proc.
Open Scope proc_scope.

Inductive action (A : Type) : Type :=
| IdleAct of A = unit
| AllocAct of A = ptr & dynamic
| DeallocAct of A = unit & ptr
| ReadAct of ptr
| WriteAct of A = unit & ptr & dynamic
(* need to ensure the value at pointer p and oldval are equal (i.e., same type and value) *)
(* CAS(p, oldval, newval) *)
| CASAct (B1 B2 : eqType) of A = bool & ptr & B1 & B2. 

(* introduce new definitions to avoid writing proofs of type equality *)
Definition Idle := IdleAct (@erefl Type unit).
Definition Alloc A (v : A) := AllocAct (@erefl Type ptr) (dyn v).
Definition Dealloc x := DeallocAct (@erefl Type unit) x.
Definition Read A x := ReadAct A x.
Definition Write A x (v : A) := WriteAct (@erefl Type unit) x (dyn v).
Definition CAS (B1 B2 : eqType) x (v1 : B1) (v2 : B2) := CASAct (@erefl Type bool) x v1 v2.

Implicit Arguments Alloc [A].
Implicit Arguments Read [A].
Implicit Arguments Write [A].
Implicit Arguments CAS [B1 B2].
Prenex Implicits Alloc Read Write CAS.

(*
Module DynAct.
Record dynamic_act : Type := dynact {typ : Type; act : action typ}.
End DynAct.

Notation dynamic_act := DynAct.dynamic_act.
Notation dynact := DynAct.dynact.

(* need to use dynact_injT/dynact_inj because case (injection) is broken, ugly *)
Lemma dynact_injV A (x y : action A) : dynact x = dynact y -> x = y.
Proof. move=>A x y [H]; apply: inj_pairT2 H. Qed.

Lemma dynact_eta d : d = dynact (DynAct.act d).
Proof. by case. Qed.

Lemma dynact_injT A1 A2 (x1 : action A1) (x2 : action A2) : 
        dynact x1 = dynact x2 -> A1 = A2.
Proof. by move=>A1 A2 x1 x2 [H]. Qed.

Lemma dynact_inj A1 A2 (a1 : action A1) (a2 : action A2) : 
        dynact a1 = dynact a2 -> A1 = A2 /\ jmeq action a1 a2.
Proof.
by move=>A1 A2 a1 a2 [E]; rewrite -{A2}E in a2 *; move/inj_pairT2=>->.
Qed.
*)

(* New case analysis on actions. Instead of *Act things which require
passing in the proof of type equality, the following case analysis is
in terms of Idle, Alloc, Dealloc, Read and Write *)

CoInductive action_spec A : action A -> Type :=
| idle_action_spec : 
    A = unit -> forall pf, action_spec (coerce Idle pf)
| alloc_action_spec : 
    A = ptr -> forall B (v : B) pf, action_spec (coerce (Alloc v) pf)
| dealloc_action_spec : 
    A = unit -> forall x pf, action_spec (coerce (Dealloc x) pf)
| read_action_spec x : action_spec (Read x)
| write_action_spec :
    A = unit -> forall B x (v : B) pf, action_spec (coerce (Write x v) pf)
| cas_action_spec :
    A = bool -> forall (B1 B2 : eqType) x (v1 : B1) (v2 : B2) pf, 
      action_spec (coerce (CAS x v1 v2) pf).

Lemma actP A : forall a : action A, action_spec a.
Proof. 
move=>[pf|pf d|pf p|p|pf p d|B1 B2 pf p v1 v2].
- rewrite (_ : IdleAct pf = coerce Idle (sym_eq pf)). 
  - by apply: idle_action_spec. 
  by move: (pf); rewrite pf; apply: Streicher_K.
- rewrite (_ : AllocAct pf d = coerce (Alloc (Dyn.val d)) (sym_eq pf)).
  - by apply: alloc_action_spec.
  by move: (pf); rewrite pf (dyn_eta d); apply: Streicher_K.
- rewrite (_ : DeallocAct pf p = coerce (Dealloc p) (sym_eq pf)).
  - by apply: dealloc_action_spec.
  by move: (pf); rewrite pf; apply: Streicher_K.
- by apply: read_action_spec.
- rewrite (_ : WriteAct pf p d = coerce (Write p (Dyn.val d)) (sym_eq pf)).
  - by apply: write_action_spec.
  by move: (pf); rewrite pf (dyn_eta d); apply: Streicher_K.
rewrite (_ : CASAct pf p v1 v2 = coerce (CAS p v1 v2) (sym_eq pf)).
- by apply: cas_action_spec.
by move: (pf); rewrite pf; apply: Streicher_K. 
Qed.

Definition Idleb A : pred (action A) := 
  [fun a => if a is IdleAct _ then true else false].

CoInductive idle_spec A : action A -> Type :=
| idle_idle_spec : A = unit -> forall pf, idle_spec (coerce Idle pf)
| ow_idle_spec (a : action A) : ~~ Idleb a -> idle_spec a.

Lemma IdleP A (a : action A) : idle_spec a.
Proof.
case/actP:a; try by move=>-> *; rewrite eqc; apply: ow_idle_spec.
- by move=>-> pf; apply: idle_idle_spec.
by move=>x; apply: ow_idle_spec.
Qed.

(*
Lemma IdlebPD A (a : action A) : 
        reflect (dynact a = dynact Idle) (Idleb a).
Proof.
move=>A; case/actP; try by move=>->*; rewrite eqc; constructor.
by move=>x; constructor.
Qed.

Lemma IdlebP A (a : action A) : 
        reflect (A = unit /\ jmeq action a Idle) (Idleb a).
Proof.
move=>A a; apply: (iffP (IdlebPD a)); last first.
- by case=>E; rewrite {A}E in a *; move/jmE=>->.
move=>E; case/dynact_injT: (E)=>T. rewrite {A}T in a E *.
by case/dynact_inj: E.
Qed.
*)

(* alternative proof, which deesn't use dynact, so we *)
(* can perhaps erase that definition *)
(*
Lemma IdlebP A (a : action A) : 
        reflect (A = unit :> Type /\ jmeq action a Idle) (Idleb a).
Proof.
move=>A; case/actP.
- by move=>-> pf; rewrite eqc; constructor.
- move=>B v -> pf; rewrite eqc; constructor; case=>E _.
  suff: exists x1 : (unit : Type), exists x2 : (unit : Type), x1 <> x2.
  - by case=>[[]][[]].
  by move: (unit : Type) E=>_ <-; exists null; exists null .+ 1.
- by move=>x -> pf; rewrite eqc; constructor; case=>_; move/jmE.
- by move=>x; constructor; case=>->; move/jmE.
by move=>B x v -> pf; rewrite eqc; constructor; case=>_; move/jmE.
Qed.
*)

(***************)
(* Transitions *)
(***************)

(* Transition is an (action, value) pair. *)
(* Used for stepping and trace semantics. *)

Module Tr.
Inductive trans : Type := 
  Transition {typ : Type; act : action typ; val : typ}.

Lemma inj_val A1 A2 (a1 : action A1) (a2 : action A2) (x1 : A1) (x2 : A2) : 
        Transition a1 x1 = Transition a2 x2 -> 
        [/\ A1 = A2, jmeq action a1 a2 & x1 =jm x2].
Proof.
by case=>E; rewrite -{}E in a2 x2 *; do 2![move/inj_pair2=>->].
Qed.

Lemma inj_valE A (a1 a2 : action A) (x1 x2 : A) : 
        Transition a1 x1 = Transition a2 x2 -> a1 = a2 /\ x1 = x2.
Proof.
by move=>H; case: (inj_val H)=>_; do 2![move/jmE=>->].
Qed.

End Tr.

Prenex Implicits Tr.inj_val Tr.inj_valE.

Notation trans := Tr.trans.

Notation "[ 'tr' A , V ]" := (Tr.Transition A V)
  (at level 0, format "'[hv' [ 'tr'  A ,  V ] ']'") : proc_scope.
Notation "[ 'idle' ]" := (Tr.Transition Idle tt) : proc_scope.
Notation "[ 'alloc' v , x ]" := (Tr.Transition (Alloc v) x) 
  (at level 0, format "'[hv' [ 'alloc'  v ,  x ] ']'") : proc_scope.
Notation "[ 'dealloc' x ]" := (Tr.Transition (Dealloc x) tt) 
  (at level 0, format "'[hv' [ 'dealloc'  x ] ']'") : proc_scope.
Notation "[ 'read' x , v ]" := (Tr.Transition (Read x) v) 
  (at level 0, format "'[hv' [ 'read'  x ,  v ] ']'") : proc_scope.
Notation "[ 'write' x , v ]" := (Tr.Transition (Write x v) tt) 
  (at level 0, format "'[hv' [ 'write'  x ,  v ] ']'") : proc_scope.
Notation "[ 'cas' x , v1 , v2 , b ]" := (Tr.Transition (CAS x v1 v2) b) 
  (at level 0, format "'[hv' [ 'cas' x , v1 , v2 , b ] ']'") : proc_scope.

Lemma tr_eta t : t = [tr Tr.act t, Tr.val t].
Proof. by case:t=>T a v. Qed.

(* case analysis on transitions *)
CoInductive trans_spec : trans -> Type := 
| idle_trans_spec : trans_spec [idle]
| alloc_trans_spec A x (v : A) : trans_spec [alloc v, x]
| dealloc_trans_spec x : trans_spec [dealloc x]
| read_trans_spec (A : Type) x (v : A) : trans_spec [read x, v]
| write_trans_spec A x (v : A) : trans_spec [write x, v]
| cas_trans_spec (B1 B2 : eqType) x (v1 : B1) (v2 : B2) b : 
    trans_spec [cas x, v1, v2, b].

Hint Constructors trans_spec.

Lemma trP : forall t, trans_spec t.
Proof.
case=>A; case/actP=>[->|-> B v|-> x||-> B x v|-> B1 B2 x v1 v2] pf;
by rewrite ?eqc //; case.
Qed.

Lemma trD1 : (forall A B x y (v : A) (w : B), 
                [read x, v] = [read y, w] <-> [/\ A = B, x = y & v =jm w]) *
             (forall A B x y (v : A) (w : B), 
                [write x, v] = [write y, w] <-> [/\ A = B, x = y & v =jm w]) *
(*             (forall (B1 B1' B2 B2' : eqType) x x' (v1 : A) (v1' : A') (v2 : B) (v2' : B') b b',
                [cas x, v1, v2, b] = [cas x', v1', v2', b'] <->
                [/\ x = x', A = A', v1 =jm v1', B = B', v2 =jm v2' & b = b']) **)
             (forall A B x y (v : A) (w : B), 
                [alloc v, x] = [alloc w, y] <-> [/\ A = B, x = y & v =jm w]) *
             (forall x y, 
                [dealloc x] = [dealloc y] <-> x = y).
Proof.
split; last by [move=>x y; split=>[[]|->]].
split=>//; last first. 
- move=>A B x y v w; split; [case/Tr.inj_valE|]; 
  case=>E; rewrite -{B}E in w *;
  by [move/inj_pair2=>->-> | move=>->; move/jmE=>->].
(*split; last first.
- move=>A A' B B' x x' v1 v1' v2 v2' b b'; split.
  - by case/Tr.inj_valE; case=>->; rbep1 v1'; rbep1 v2'=>->.
  by case=>-> []; rbejm1 v1'=>[[]]; rbejm1 v2'=>->.*)
split; last first.
- move=>A B x y v w; split. 
  - by case=>-> E; rewrite -{B}E in w *; move/inj_pair2=>->.
  by case=>E; rewrite -{B}E in w *; move=>->; move/jmE=>->.
move=>A B x y v w; split; case=>E ->; rewrite -{B}E in w *;
by [move/inj_pair2=>-> | move/jmE=>->].
Qed.

(* XXX should the equality for cas appear here? *)
Lemma trD2 : (forall A x y (v w : A), [read x, v] = [read y, w] <-> 
                                      x = y /\ v = w) *
             (forall A x y (v w : A), [write x, v] = [write y, w] <-> 
                                      x = y /\ v = w) *
             (forall A x y (v w : A), [alloc v, x] = [alloc w, y] <->
                                      x = y /\ v = w).
Proof. 
split; [split|]; move=>A x y v w; rewrite trD1; 
by (split; [case=>_ ->; move/jmE=>->|case=>->->]).
Qed.

Lemma trD3 : forall A1 A2 a1 a2 (x1 : A1) (x2 : A2), 
               [tr a1, x1] = [tr a2, x2] <-> 
               [/\ A1 = A2, jmeq action a1 a2 &
                   forall pf, coerce (T:=id) x1 pf = x2].
Proof.
move=>A1 A2 a1 a2 x1 x2; split; case=>E; rewrite -{A2}E in a2 x2 *. 
- by do 2![move/inj_pair2=>->]; split=>// pf; rewrite eqc.
by move/jmE=>->; move/(_ (erefl _))=>/= ->.
Qed.

Lemma trD4 : forall A a1 a2 (x1 x2 : A), 
               [tr a1, x1] = [tr a2, x2] <-> a1 = a2 /\ x1 = x2.
Proof.
move=>*; rewrite trD3.
split; last by case=>->->; split=>// pf; rewrite eqc.
by case=>_; move/jmE=>->; move/(_ (erefl _))=>/= ->.
Qed.

(* trD2 and trD4 must appear before trD1 and trD3, respectivelly *)
Definition trD := (trD2, trD1, trD4, trD3). 

Definition idleb : pred trans := 
  [fun t => if Tr.act t is IdleAct _ then true else false].

Lemma idleP t : reflect (t = [idle]) (idleb t).
Proof. by case/trP:t; rewrite /idleb /=; constructor. Qed.

(***************)
(* Annotations *)
(***************)

(*
Section Annotations.
Variables (U : pcm).

Definition annot_axiom (p : Pred U) := 
  forall x y, x \In p -> x \+ y \In p.
 
Structure annot := Annot {
  assert : Pred U; 
  next : U; 
  _ : annot_axiom assert}. 

Lemma id_annot_mono : annot_axiom PredT.
Proof. by []. Qed. 

Definition id_annot := @Annot PredT Unit id_annot_mono.

Lemma assertP x y f : x \In assert f -> x \+ y \In assert f.
Proof. by case: f=>f z; apply. Qed.

End Annotations.
*)

Section Annotations.
Variables (U : pcm).

Definition annot_axiom (f : U -> option U) := 
  forall x x' y,
    f x = Some x' ->
    valid (x \+ y) ->
    f (x \+ y) = Some (x' \+ y).

Structure annot := Annot {
  assert : U -> option U;
  _ : annot_axiom assert}.

Definition idopt (x : U) := Some x.
Lemma id_annot_mono : annot_axiom idopt.
Proof. by move=>x x' y [<-{x'}]. Qed. 

Definition id_annot := @Annot idopt id_annot_mono.

Lemma assertP x x' y f :
        assert f x = Some x' ->
        valid (x \+ y) ->
        assert f (x \+ y) = Some (x' \+ y).
Proof. by case: f=> f H; apply: H. Qed.
End Annotations.

Implicit Arguments id_annot [U].
Prenex Implicits id_annot.

(*********)
(* Trees *)
(*********)

Section Trees.
Variables U : pcm.

(* trees with annotations on Cons's *)

Inductive tree A :=
  Dummy | Ret of A | Cons B of action B & annot U & B -> tree A |
  Par B1 B2 of tree B1 & tree B2 & B1 * B2 -> tree A.

Implicit Arguments Dummy [A].
Prenex Implicits Dummy.

Implicit Arguments Ret [A].
Prenex Implicits Ret.

Implicit Arguments Par [A B1 B2].
Prenex Implicits Par.

Definition Retb A : pred (tree A) :=
  [fun a => if a is Ret _ then true else false].

CoInductive ret_spec A : tree A -> Type :=
| dummy_ret_spec : ret_spec (Dummy)
| ret_ret_spec (v : A) : ret_spec (Ret v)
| cons_ret_spec B a f (p : B -> tree A) : ret_spec (Cons a f p)
| par_ret_spec B1 B2 p1 p2 (p : B1 * B2 -> tree A) : ret_spec (Par p1 p2 p).

Lemma RetP A (p : tree A) : ret_spec p.
Proof.
case:p; move=>*;
by [apply: dummy_ret_spec|apply: ret_ret_spec
   |apply: cons_ret_spec|apply: par_ret_spec].
Qed.

Lemma RetbP A (p : tree A) : 
        reflect (exists v, p = Ret v) (Retb p).
Proof.
case:p; try by move=>*; constructor; case=>v.
by move=>v; constructor; exists v.
Qed.

(* equal conses have equal head actions and continuations *)
Lemma consE (A1 A2 : Type) B (a1 : action A1) (a2 : action A2) t1 t2 g1 g2 : 
        Cons a1 g1 t1 = Cons a2 g2 t2 :> tree B <-> 
          [/\ A1 = A2, jmeq action a1 a2, g1 = g2 &
              jmeq (fun A => A -> tree B) t1 t2].
Proof. 
split; case=>E; rewrite -{A2}E in a2 t2 *.
- by move/inj_pair2=>->->; move/inj_pair2=>->. 
by move/jmE=>->->; move/jmE=>->.
Qed.

Lemma consEX (A B : Type) (a1 a2 : action A) t1 t2 g1 g2 : 
        Cons a1 g1 t1 = Cons a2 g2 t2 :> tree B <-> 
        [/\ a1 = a2, g1 = g2 & t1 = t2].
Proof. 
split; case; first by rbp=>->; rbp.
by move=>->->->.
Qed.

(* similarly for pars *)
Lemma parE (A1 B1 A2 B2 C : Type) (p1 : tree A1) (p2 : tree A2) 
            (q1 : tree B1) (q2 : tree B2) r1 r2 : 
        Par p1 q1 r1 = Par p2 q2 r2 :> tree C <-> 
          [/\ A1 = A2, B1 = B2, jmeq tree p1 p2, 
              jmeq tree q1 q2 &
              jmeq2 (fun A B => A * B -> tree C) r1 r2].
Proof. 
split; case=>E1 E2; rewrite -{A2}E1 -{B2}E2 in p2 q2 r1 r2 *.
- do 2![move/inj_pair2=>->].
  by do 2![move/inj_pair2]; move=>->.
by do 2![move/jmE=>->]; move/jm2E=>->.
Qed.

Lemma parEX (A1 A2 C : Type) (p1 p2 : tree A1) (q1 q2 : tree A2) r1 r2 :
        Par p1 q1 r1 = Par p2 q2 r2 :> tree C <-> 
          [/\ p1 = p2, q1 = q2 & r1 = r2].
Proof. 
split; case; first by rbp; rbp; move/inj_pair2; rbp.
by move=>->->->.
Qed.

(**********************************)
(* Some basic operations on trees *)
(**********************************)

(* these should go in prelude or something *)
Lemma contV B (P : B -> B -> Prop) :
        (forall (x x' : B), x =jm x' -> P x x') <-> forall x, P x x.
Proof.
split.
- by move=>H x; exact: (H x x (jmeq_refl _)).
by move=>H x x'; move/jmE=>->.
Qed.

Lemma contVT B (P : B -> B -> Prop) :
        (forall (x x' : B), B = B -> x =jm x' -> P x x') <-> forall x, P x x.
Proof.
split.
- by move=>H x; exact: (H x x (erefl _) (jmeq_refl _)). 
by move=>H x x' _; move/jmE=>->.
Qed.

Fixpoint tmap A B (f : A -> B) (t : tree A) :=
  match t with 
    Dummy => Dummy
  | Ret x => Ret (f x)
  | Cons _ a g t' => Cons a g (fun x => tmap f (t' x))
  | Par _ _  t1 t2 t' => Par t1 t2 (fun x => tmap f (t' x))
  end.

Lemma par_compM T A1 A2 (p1 : tree A1) (p2 : tree A2) (f : A1 * A2 -> T) :
        Par p1 p2 (Ret \o f) = tmap f (Par p1 p2 Ret).
Proof. by []. Qed.

Definition to_tree A (x : option A) :=
  match x with
  | Some y => Ret y
  | None => Dummy
  end.

Definition to_tree1 A A' (f : A -> option A') x :=
  match f x with
  | Some y => Ret y
  | None => Dummy
  end.

Fixpoint tcat A B (p : tree A) (q : A -> tree B) :=
  match p with
    Dummy => Dummy
  | Ret a => q a
  | Cons _ t g r => Cons t g (fun x => tcat (r x) q)
  | Par _ _ t1 t2 r => Par t1 t2 (fun x => tcat (r x) q)
  end.

Fixpoint tprefix' A A' (p : tree A) (p' : tree A') : Prop :=
  match p, p' with
  | Dummy, _ => True
  | Ret x, Ret x' => x =jm x'
  | Cons B a g p, Cons B' a' g' p' =>
    [/\ B = B',
        jmeq action a a',
        g = g' 
      & forall x x', x =jm x' -> tprefix' (p x) (p' x')]
  | Par B1 B2 p1 p2 p, Par B1' B2' p1' p2' p' =>
    [/\ B1 = B1',
        B2 = B2',
        tprefix' p1 p1',
        tprefix' p2 p2'
        & forall x x', x =jm x' -> tprefix' (p x) (p' x')]
  | _, _ => False
  end.

Definition tprefix A (p p' : tree A) : Prop := tprefix' p p'.

Notation "p << q" := (tprefix p q) (at level 70).

Lemma tpfx_refl A (p : tree A) : p << p.
Proof.
elim:p=>{A} //.
- by move=>A a; apply jmeq_refl.
- by move=>A B a g p IH; split=>// b b'; move/jmE=><-; apply:(IH b).
move=>A B B' t1 IH1 t2 IH2 t' IH'; split=>// b b';
by move/jmE=><- {b'}; apply:IH' b.
Qed.

Hint Resolve tpfx_refl.

Lemma tpfx_trans A (p p' p'' : tree A) : p << p' -> p' << p'' -> p << p''.
Proof.
elim:p p' p''=>{A} // A.
- by move=>v [] // v' [] // v''; do 2!move/jmE=>->.
- move=>B a g p IH [] // B' a' g' p' [] // B'' a'' g'' p''.
  case; rbejm2 a' p'; move=>->; move/contV=>Ep'.
  case; rbejm2 a'' p''; move=>->; move/contV=>Ep'';
  by split=>//; apply/contV=>x; apply: IH _ _ _ (Ep' x) (Ep'' x).
move=>B1 B2 p1 IH1 p2 IH2 p IH [] // B1' B2' p1' p2' p'
      [] // B1'' B2'' p1'' p2'' p'' [];
rbe2 p1' p'; rbe2 p2' p'; move=>L1 L2; move/contV=>L' [];
rbe2 p1'' p''; rbe2 p2'' p''; move=>L1' L2'; move/contV=>L''; split=>//;
by [apply: (IH1 _ _  L1 L1')
   |apply: (IH2 _ _  L2 L2')
   |apply/contV=>b; apply: IH _ _ _ (L' b) (L'' b)].
Qed.

Definition tprefix1 A C (p p' : C -> tree A) : Prop := forall x, p x << p' x.

Notation "p <<1 q" := (tprefix1 p q) (at level 70).

Lemma tpfxP A (p p' : tree A) :
  p' << p <->
  [\/ p' = Dummy,
      exists x, p' = Ret x /\ p = Ret x,
      exists C, exists a : action C, exists g, exists r', exists r,
        [/\ p' = Cons a g r', p = Cons a g r & r' <<1 r]
    | exists B1, exists B2, exists p1', exists p2', exists r' : B1 * B2 -> tree A,
      exists p1, exists p2, exists r,
        [/\ p' = Par p1' p2' r', p = Par p1 p2 r,
            p1' << p1, p2' << p2 & r' <<1 r]].
Proof.
split.
- case:p'.
  - by move=>*; apply: Or41.
  - by move=>v'; case:p=>// v; move/jmE=>->; apply: Or42; exists v.
  - move=>B' a' g' r'; case:p=>// B a g r []; rfejm2 a' r'; move=>->.
    by move/contV=>E; apply: Or43; exists B; exists a; 
       exists g; exists r'; exists r.
  move=>B1' B2' p1' p2' r'; case:p=>// B1 B2 p1 p2 r [];
  rfe2 p1' r'; rfe2 p2' r'=>L1 L2; move/contV=>L; apply: Or44;
  exists B1; exists B2; exists p1'; exists p2'; exists r';
  by exists p1; exists p2; exists r.
case=>[-> | [v] [->->] | [B][a][g][r'][r] [->->L]
      | [B1][B2][p1'][p2'][r'][p1][p2][r] [->->L1 L2 L]] //;
by split=>//; apply/contV.
Qed.

Lemma tpfx_XdummyP A (p' : tree A) :
  p' << @Dummy A -> p' = Dummy.
Proof. by case:p'=>//. Qed.

Lemma tpfx_retXP A (p : tree A) x' :
  Ret x' << p -> p = Ret x'.
Proof. by case:p x'=>// v v'; move/jmE=>->. Qed.

Prenex Implicits tpfx_retXP.

Lemma tpfx_XretP A (p' : tree A) x :
  p' << Ret x -> p' = Dummy \/ p' = Ret x.
Proof.
case:p' x=>//; first by move=>*; left.
by move=>v v'; move/jmE=>->; right.
Qed.

Prenex Implicits tpfx_XretP.

Lemma tpfx_retP A (x x' : A) :
  Ret x' << Ret x <-> x' = x.
Proof. split; first by move/jmE=>->. by move=>->. Qed.

Lemma tpfx_consXP A (p : tree A) C (a : action C) g r' :
        Cons a g r' << p -> exists r, p = Cons a g r /\ r' <<1 r.
Proof.
case: p C a g r'=>// B a g r C a' g' r' []; rfejm2 a' r'; move=>->.
by move/contV=>E; exists r.
Qed.

Lemma tpfx_XconsP A (p' : tree A) C (a : action C) g r :
        p' << Cons a g r -> 
        p' = Dummy \/ exists r', p' = Cons a g r' /\ r' <<1 r.
Proof.
case: p' C a g r=>//; first by move=>*; left.
move=>B a g r C a' g' r' []; rbejm2 a' r'; move=>->.
by move/contV; right; exists r.
Qed.

Lemma tpfx_consPT A B B' (a : action B) (a' : action B') g g'
                         (r : B -> tree A) (r' : B' -> tree A) :
        Cons a' g' r' << Cons a g r <-> 
        [/\ B = B', jmeq action a a', g = g' & 
            forall x x', x =jm x' -> r' x << r x'].
Proof.
split; last by case; rbe2 a' r'; rbjm=>L.
case/tpfxP=>[|[v]|[_B][_a][_g][_r][_r']|
             [_A][_B][_p][_q][_r][_p'][_q'][_r']] // [] //.
case/consE; rfe2 a' r'; rbjm=>->; rbjm. 
case/consE; rfe2 a r; rbjm=>->; rbjm. 
by move=>H; split=>//; apply/contV.
Qed.

Lemma tpfx_consP A C a a' (r r' : C -> tree A) g g' :
        Cons a' g' r' << Cons a g r <-> 
        [/\ a = a', g = g' & r' <<1 r].
Proof.
split; first by case/tpfx_consPT=>_; rbjm=>->; move/contV.
by case=>->-> L; apply/tpfx_consPT; split=>//; apply/contV.
Qed.

Lemma tpfx_parXP A (p : tree A) B1 B2 p1' p2' (r' : B1 * B2 -> tree A) :
        Par p1' p2' r' << p ->
        exists p1, exists p2, exists r,
          [/\ p = Par p1 p2 r, p1' << p1, p2' << p2 & r' <<1 r].
Proof.
case:p B1 B2 p1' p2' r'=>// B1 B2 p1 p2 r B1' B2' p1' p2' r' [].
rfe2 p1' r'; rfe2 p2' r'; move=>L1 L2; move/contV=>L.
by exists p1; exists p2; exists r.
Qed.

Lemma tpfx_XparP A (p' : tree A) B1 B2 p1 p2 (r : B1 * B2 -> tree A) :
  p' << Par p1 p2 r ->
  p' = Dummy \/
  exists p1', exists p2', exists r',
    [/\ p' = Par p1' p2' r', p1' << p1, p2' << p2 & r' <<1 r].
Proof.
case:p' B1 B2 p1 p2 r=>// [|B1' B2' p1' p2' r'] B1 B2 p1 p2 r; first by left.
case; rfe2 p1' r'; rfe2 p2' r'; move=>L1 L2; move/contV=>L;
by right; exists p1'; exists p2'; exists r'.
Qed.

Lemma tpfx_parP A B1 B2 p1 p1' p2 p2' (r r' : B1 * B2 -> tree A) :
  Par p1' p2' r' << Par p1 p2 r <->
  [/\ p1' << p1, p2' << p2 & r' <<1 r].
Proof.
split.
- by case=>_ _ L1 L2; move/contV=>L.
by case=>L1 L2 L; split=>//; apply/contV.
Qed.


(*
(* catenating a tree with a set *)

Fixpoint tcatp A B (t1 : tree A) (ts : A -> Pred (tree B)) : Pred (tree B) :=
  fun t2 => 
    match t1 with
    | Dummy => t2 = Dummy
    | Ret x => t2 \In ts x
    | Cons C a r1 =>
        exists r2, t2 = a @ r2 /\ forall x, r2 x \In tcatp (r1 x) ts
    | Par B1 C1 p q r1 =>
        exists r2, t2 = Par p q r2 /\ forall x, r2 x \In tcatp (r1 x) ts
    end.

(* a better case-analysis, to avoid the existentials in *)
(* Cons and Par cases *)

CoInductive tcatp_lspec A B (ts : A -> Pred (tree B)) (t2 : tree B) : 
                        tree A -> Prop :=
| catp_lspec_dummy of t2 = Dummy : tcatp_lspec ts t2 Dummy
| catp_lspec_ret x of t2 \In ts x : tcatp_lspec ts t2 (Ret x)
| catp_lspec_cons C (a : action C) r1 r2 of
    t2 = a @ r2 & (forall x, r2 x \In tcatp (r1 x) ts) : 
    tcatp_lspec ts t2 (a @ r1)
| catp_lspec_par B1 C1 p q (r1 r2 : B1 * C1 -> _) of
  t2 = Par p q r2 & (forall x, r2 x \In tcatp (r1 x) ts) : 
    tcatp_lspec ts t2 (Par p q r1).

Lemma tcatpL A B (t1 : tree A) ts (t2 : tree B) : 
        t2 \In tcatp t1 ts <-> tcatp_lspec ts t2 t1.
Proof.
split; last first.
- case=>//=; first by move=>C a r1 r2 ->; exists r2.
  by move=>B1 C1 p q r1 r2 ->; exists r2.
case: t1=>/=.
- by apply: catp_lspec_dummy.
- by move=>a H; apply: catp_lspec_ret.
- by move=>C a r1 [r2][->]; apply: catp_lspec_cons.
by move=>B1 B2 p q r1 [r2][->]; apply: catp_lspec_par.
Qed.

Lemma tcatp_dummyP A B p : @Dummy A \In tcatp (@Dummy B) p.
Proof. by []. Qed.

Lemma tcatp_XdummyP A B (q12 : tree B) (x2 : A -> Pred (tree B)) :
        q12 \In tcatp (@Dummy A) x2 <-> q12 = Dummy.
Proof. by split; [case=>-> | move=>->; apply/tcatp_dummyP]. Qed.

Lemma tcatp_retP A B (t : tree A) (x : B) p :
        Ret x \In tcatp t p <->
        exists y, t = Ret y /\ Ret x \In p y.
Proof.
rewrite tcatpL; split; first by case=>// y H; exists y.
by case=>y [->] H; apply: catp_lspec_ret.
Qed.

Lemma tcatp_XretP A B (v : A) (q12 : tree B) (x2 : A -> Pred (tree B)) :
        q12 \In tcatp (Ret v) x2 <-> q12 \In x2 v.
Proof. by []. Qed.

Lemma tcatp_XconsP A B C a (r1 : A -> tree B) ts (t2 : tree C) :
        t2 \In tcatp (a @ r1) ts <->
        exists r2, t2 = a @ r2 /\ forall v, r2 v \In tcatp (r1 v) ts.
Proof. by []. Qed.

Definition tcatp_consP := tcatp_XconsP.

Lemma tcatp_XparP A B B1 C1 p q (r1 : B1 * C1 -> tree A) (t2 : tree B) ts :
        t2 \In tcatp (Par p q r1) ts <->
        exists r2, t2 = Par p q r2 /\ forall x, r2 x \In tcatp (r1 x) ts.
Proof. by []. Qed.

Definition tcatp_parP := tcatp_XparP.
*)

Definition tcatp A1 A2 (q1 : tree A1) (x2 : A1 -> Pred (tree A2)) :=
  let fix tcatp t1 t' : Prop :=
      match t1, t' with
      | Dummy, Dummy => True
      | Ret v, _ => t' \In x2 v
      | Cons C a g r, Cons C' a' g' r' =>
        [/\ C = C',
            jmeq action a a', g = g'
            & forall x x', x =jm x' -> tcatp (r x) (r' x')]
      | Par C1 C2 t1 t2 r, Par C1' C2' t1' t2' r' =>
        [/\ C1 = C1',
            jmeq tree t1 t1',
            C2 = C2',
            jmeq tree t2 t2'
          & forall x x', x =jm x' -> tcatp (r x) (r' x')]
      | _, _ => False
      end
  in [Pred q12 | tcatp q1 q12].

Lemma tcatpP A1 A2 (q1 : tree A1) (q12 : tree A2) x2 :
        q12 \In tcatp q1 x2 <->
        [\/ q1 = Dummy /\ q12 = Dummy,
            exists v1, q1 = Ret v1 /\ q12 \In x2 v1,
            exists C, exists a, exists g, exists r : C -> tree A1,
              q1 = Cons a g r /\
              exists r' : C -> tree A2, q12 = Cons a g r' /\
                forall x : C, (r' x) \In tcatp (r x) x2
          | exists C1, exists C2, exists t1, exists t2, exists r : C1 * C2 -> tree A1,
              q1 = Par t1 t2 r /\
              exists r' : C1 * C2 -> tree A2, q12 = Par t1 t2 r' /\
                forall x, (r' x) \In tcatp (r x) x2].
Proof.
split=>//.
- case:q1=>[|v|B a g p|B1 B2 p1 p2 p] // H;
  [apply: Or41|apply: Or42|apply: Or43|apply: Or44].
  - by case: q12 H=>//.
  - by exists v.
  - case: q12 H=>// B' a' g' p' []; rbejm2 a' p'; move=>->; move/contV=>E;
    by exists B; exists a; exists g'; exists p; split=>//; exists p'.
  - case:q12 H=>// B1' B2' p1' p2' p' [];
    rbejm2 p1' p'; rbejm2 p2' p'; move/contV=>E;
    by exists B1; exists B2; exists p1; exists p2; exists p; split=>//; exists p'.
case:q1=>[|v|B a g p|B1 B2 p1 p2 p];
case=>[|[v1]|[B'][a'][g'][p']|[B1'][B2'][p1'][p2'][p']] [] //.
- by case: q12=>//.
- by case=>->.
- case; rbe2 a' p'; move/inj_pair2=><-<-; move/inj_pair2=><-.
  by case=>[p''] [-> E]; split=>//; apply/contV.
case; rbe2 p1' p'; rbe2 p2' p'; do 2!move/inj_pair2=><-;
by (do 2!move/inj_pair2)=><- [p''] [-> E]; split=>//; apply/contV.
Qed.

Lemma tcatp_dummyP A1 A2 x2 : @Dummy A2 \In tcatp (@Dummy A1) x2.
Proof. by []. Qed.

Lemma tcatp_XdummyP A1 A2 (q12 : tree A2) (x2 : A1 -> Pred (tree A2)) :
        q12 \In tcatp (@Dummy A1) x2 <-> q12 = Dummy.
Proof. 
split; first by case/tcatpP=>[|[v1]|[B'][g'][a'][p']|
                              [B1'][B2'][p1'][p2'][p']] []//.
move=>->; apply/tcatp_dummyP.
Qed.

(*
Lemma tcatp_dummyXP A1 A2 (q1 : tree A1) (x2 : A1 -> Pred (tree A2)) :
  (@Dummy A2) \In tcatp q1 x2.
Proof. 
case:q1=>[|v|B a p|A B p q r].
- by apply/tcatp_dummyP.
- apply/tcatpP.

apply: Or42; exists v.

split; first by case/tcatpP=>[|[v1]|[B'][a'][p']|[B1'][B2'][p1'][p2'][p']] []//.
move=>->; apply/tcatp_dummyP.
Qed.

Lemma tcatp_XretP A1 A2 (v1 : A1) (q12 : tree A2) (x2 : A1 -> Pred (tree A2)) :
  q12 \In tcatp (Ret v1) x2 <-> q12 = Dummy \/ q12 \In x2 v1.
Proof. 
split.
- by case/tcatpP=>[|[v]|[_][_][_]|[_][_][_][_][_]] [] //; case=>-> H; right.
case.
- move=>->.

split; first by case/tcatpP=>[|[v1]|[B'][a'][p']|[B1'][B2'][p1'][p2'][p']] []//.
move=>->; apply/tcatp_dummyP.
Qed.
*)

Lemma tcatp_XretP A1 A2 (v1 : A1) (q12 : tree A2) (x2 : A1 -> Pred (tree A2)) :
        q12 \In tcatp (Ret v1) x2 -> q12 \In x2 v1.
Proof. 
by case/tcatpP=>[|[v]|[_][_][_][_]|[_][_][_][_][_]] [] //; case=>-> H.
Qed.

Lemma tcatp_retP A1 A2 (q1 : tree A1) (v2 : A2) x2 :
        (Ret v2) \In tcatp q1 x2 <->
        exists v1, q1 = Ret v1 /\ Ret v2 \In x2 v1.
Proof.
split.
- case/tcatpP=>[|[v1]|[B][a][g][p][_][p']|
     [B1][B2][p1][p2][p][_][p']] [] // -> H;
  by exists v1.
by case=>v1 [->H]; apply/tcatpP; apply: Or42; exists v1.
Qed.

Lemma tcatp_consP A1 A2 B (a : action B) (q1' : B -> tree A1) 
                  (q12 : tree A2) g x2 :
        q12 \In tcatp (Cons a g q1') x2 <->
        exists q12' : B -> tree A2, q12 = Cons a g q12' /\
          forall x : B, (q12' x) \In tcatp (q1' x) x2.
Proof.
split.
- case/tcatpP=>[|[v1]|[B'][a'][g'][p']|[B1'][B2'][p1'][p2'][p']] [] // [].
  by rbe2 a' p'; move/inj_pair2=><-<-; move/inj_pair2=><-.
by case=>p' [->E]; split=>//; apply/contV.
Qed.

Lemma tcatp_XconsP A1 A2 A3 (q12 : tree A3) a (r : A1 -> tree A2) x2 g :
        q12 \In tcatp (Cons a g r) x2 <->
        exists r', q12 = Cons a g r' /\ forall v, r' v \In tcatp (r v) x2.
Proof. 
split.
- case/tcatpP=>[|[_v]|[B'][a'][g'][p']|[_A][_B][_p][_q][_r]] [] //.
  by case/consE; rbe2 a' p'; rbjm=>->; rbjm; case=>r' [->] H; exists r'.
case=>r' [->H]; apply/tcatpP; apply: Or43; exists A1, a, g, r; split=>//;
by exists r'.
Qed.

Lemma tcatp_parP A1 A2 B1 B2 p1 p2 (p : B1 * B2 -> tree A1) 
                (q12 : tree A2) x2 :
        q12 \In tcatp (Par p1 p2 p) x2 <->
        exists p' : B1 * B2 -> tree A2, q12 = Par p1 p2 p' /\
          forall x, (p' x) \In tcatp (p x) x2.
Proof.
split.
- case/tcatpP=>[|[v_]|[B_][a_][g_][p_]|[B1'][B2'][p1'][p2'][p']] [] // [];
  rbe2 p1' p'; rbe2 p2' p';
  by do 2!move/inj_pair2=><-; (do 2!move/inj_pair2)=><-.
by case=>p' [->E]; split=>//; apply/contV.
Qed.

Prenex Implicits tcatp_XretP.

Lemma tcatp_subset A1 A2 (p1 : tree A1) (p12 : tree A2) s2 s2' :
        (forall p2 v1, p2 \In s2 v1 -> p2 \In s2' v1) ->
        p12 \In tcatp p1 s2 ->
        p12 \In tcatp p1 s2'.
Proof.
move=>S.
elim: p1 p12 s2 s2' S=>[A|A v|A C a g r IH|
                        A C1 C2 p1 IH1 p2 IH2 r IH] p12 s2 s2' S.
- by move/tcatp_XdummyP=>->.
- by move/tcatp_XretP/S.
- case/tcatp_XconsP=>r' [->H]; apply/tcatp_consP.
  by exists r'; split=>//v; apply:IH (H v).
case/tcatp_parP=>r' [->H]; apply/tcatp_parP.
by exists r'; split=>//v; apply:IH (H v).
Qed.

Prenex Implicits tcatp_XretP.

(* q12 \In q1 x2 iff q12 is a prefix of q1 or (q1 ; x2) *)
Definition tpfxcat A1 A2 (q1 : tree A1) (x2 : A1 -> Pred (tree A2)) :=
  let fix tpfxcat t1 t' : Prop :=
      match t1, t' with
      | _, Dummy => True
      | Ret v, _ => exists t2, t2 \In x2 v /\ t' << t2
      | Cons C a g r, Cons C' a' g' r' =>
        [/\ C = C',
            jmeq action a a', g = g'
            & forall x x', x =jm x' -> tpfxcat (r x) (r' x')]
      | Par C1 C2 t1 t2 r, Par C1' C2' t1' t2' r' =>
        [/\ C1 = C1',
            jmeq tree t1 t1',
            C2 = C2',
            jmeq tree t2 t2'
          & forall x x', x =jm x' -> tpfxcat (r x) (r' x')]
      | _, _ => False
      end
  in [Pred q12 | tpfxcat q1 q12].

Lemma tpfxcatP A1 A2 (q1 : tree A1) (q12 : tree A2) x2 :
        q12 \In tpfxcat q1 x2 <->
        [\/ q12 = Dummy,
            exists v1, q1 = Ret v1 /\ exists q2, q2 \In x2 v1 /\ q12 << q2,
            exists C, exists a, exists g, exists r : C -> tree A1,
              q1 = Cons a g r /\
              exists r' : C -> tree A2, q12 = Cons a g r' /\
                forall x : C, (r' x) \In tpfxcat (r x) x2
          | exists C1, exists C2, exists t1, exists t2,
            exists r : C1 * C2 -> tree A1,
              q1 = Par t1 t2 r /\
              exists r' : C1 * C2 -> tree A2, q12 = Par t1 t2 r' /\
                forall x, (r' x) \In tpfxcat (r x) x2].
Proof.
split=>//.
- case:q1=>[|v|B a g p|B1 B2 p1 p2 p] //.
  (* [apply: Or41|apply: Or42|apply: Or43|apply: Or44]. *)
  - by case:q12=>// H; apply: Or41.
  - (case:q12=>//; first by move=>*; apply: Or41);
    by move=>*; apply: Or42; exists v; split=>//.
  - (case:q12=>//; first by move=>*; apply: Or41);
    move=>B' a' g' p' []; rbejm2 a' p'; move=>->; move/contV=>E;
    by apply: Or43; exists B; exists a; exists g'; exists p; 
       split=>//; exists p'.
  - (case:q12=>//; first by move=>*; apply: Or41);
    move=>B1' B2' p1' p2' p' []; rbejm2 p1' p'; rbejm2 p2' p'; move/contV=>E;
    apply: Or44; exists B1; exists B2; exists p1; exists p2; exists p;
    by split=>//; exists p'.
case=>[|[v1]|[B'][a'][g'][p']|[B1'][B2'][p1'][p2'][p']] [] // ->.
- by case: q1=>//.
- by case: q12=>//; move=>-> [r'] [->H] split=>//; apply/contV.
- by move=>[r'][->H]; split=>//; apply/contV.
by move=>[r'][->H]; split=>//; apply/contV.
Qed.

Lemma tpfxcat_XdummyP A1 A2 q12 (x2 : A1 -> Pred (tree A2)) :
        q12 \In tpfxcat Dummy x2 -> q12 = Dummy.
Proof. by case:q12. Qed.

Lemma tpfxcat_dummyXP A1 A2 q1 (x2 : A1 -> Pred (tree A2)) :
        Dummy \In tpfxcat q1 x2.
Proof. by case:q1. Qed.

Lemma tpfxcat_retXP A1 A2 (q1 : tree A1) (v2 : A2) x2 :
        (Ret v2) \In tpfxcat q1 x2 <->
        exists v1, q1 = Ret v1 /\ Ret v2 \In x2 v1.
Proof.
split.
- by case:q1=>// v1 [q2] [H1 H2]; rewrite (tpfx_retXP H2) in H1; exists v1.
by case=>v1 [->H]; exists (Ret v2).
Qed.

Lemma tpfxcat_XretP A1 A2 (v1 : A1) (q12 : tree A2) x2 :
        q12 \In tpfxcat (Ret v1) x2 <->
        q12 = Dummy \/ exists q2, q2 \In x2 v1 /\ q12 << q2.
Proof.
split.
- case/tpfxcatP=>[|[v][[<-]H]|[B][a][g][p][]|[B1][B2][p1][p2][p][]] //;
  by [left | right].
case=>[->|H]; apply/tpfxcatP; by [apply: Or41 | apply: Or42; exists v1].
Qed.

Lemma tpfxcat_XconsP A1 A2 B (a : action B) 
                     (q1' : B -> tree A1) (q12 : tree A2) x2 g :
        q12 \In tpfxcat (Cons a g q1') x2 <->
        q12 = Dummy \/
        exists q12' : B -> tree A2, q12 = Cons a g q12' /\
          forall x : B, (q12' x) \In tpfxcat (q1' x) x2.
Proof.
split.
- case/tpfxcatP=>[|[v1][]|[B'][a'][g'][r'][]|[B1][B2][p1][p2][p'][]] //.
  - by move=>->; left.
  by case/consE; rbejm2 a' r'; move=>->; rbjm=>H; right.
case; first by move=>->.
by case=>q12' [->H]; split=>//; apply/contV.
Qed.

Lemma tpfxcat_XparP A1 A2 B1 B2 p1 p2 (p : B1 * B2 -> tree A1) (q12 : tree A2) x2 :
        q12 \In tpfxcat (Par p1 p2 p) x2 <->
        q12 = Dummy \/
        exists p' : B1 * B2 -> tree A2, q12 = Par p1 p2 p' /\
          forall x, (p' x) \In tpfxcat (p x) x2.
Proof.
split.
- case/tpfxcatP=>[->|[v1][]|[B][a][g][r][]|[B1'][B2'][p1'][p2'][p']] //;
  first by left.
  by case; case/parE; rbe2 p1' p'; rbe2 p2' p'; 
     do 2!rbjm; move/jm2E=><- H; right.
case; first by move=>->.
by case=>p' [->]; split=>//; apply/contV.
Qed.

End Trees.

Implicit Arguments Dummy [U A].
Implicit Arguments Ret [U A].
Implicit Arguments Par [U A B1 B2].

Prenex Implicits Dummy Ret Par.
Prenex Implicits tpfx_retXP tpfx_XretP tcatp_XretP tcatp_XretP tcatp_XretP.

Hint Resolve tpfx_refl.

Notation "a @ p" := (Cons a id_annot p)
  (at level 20, right associativity) : proc_scope.

Notation "a '@@' p" := (Cons a id_annot (fun _ => p))
  (at level 39, right associativity, 
   format "a  '/' '@@'  p") : proc_scope.

Notation "p << q" := (tprefix p q) (at level 70).
Notation "p <<1 q" := (tprefix1 p q) (at level 70).

