Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq rels ordtype.
Require Import Eqdep prelude domains heaps.
Require Import pcm resources8 proc9 step9 hstep9 sbisim9 wbisim9.
Require Import locksafe9.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(* CSL-style logic *)

Section CSL_Logic. 
Variable R : resource.

Notation U := (univ_of R).
Notation tree A := (tree U A).
Notation lk := (lock_of R).
Notation procHeap := (procHeap U).
Notation Inv := (Inv (R:=R)).

(* let's try dropping sh from the specs *)
(* as csl never talks about sh *)

Definition pre := U -> procHeap -> Prop.
Definition cont A := A -> pre.
Definition post A := A -> U -> U -> procHeap -> procHeap -> Prop.

(* maybe i can drop ep from the specs too *)

Definition spec A := (pre * post A)%type.

Definition has_spec A (s : spec A) := 
  [Pred p : tree A | 
     forall sh lp ep, 
       [sh, lp, ep] \In coherent -> 
       s.1 (gh ep) lp -> 
       lock_safe [sh, lp, ep] p /\ 
       after [sh, lp, ep] p
         (fun s' v' =>
           let: [sh', lp', ep'] := s' in 
             s.2 v' (gh ep) (gh ep') lp lp')].

Structure prog A (s : spec A) :=
  Prog {model : Pred (tree A);
        _ : model <=p has_spec s}.

Lemma specE A (s : spec A) (e : prog s) p : 
        p \In model e -> p \In has_spec s.
Proof. by case: e=>/= m H; move/H. Qed.

(*********************)
(* Lattice structure *)
(*********************)

Section Lattice.
Variables (A : Type) (s : spec A).
 
Definition pleq (e1 e2 : prog s) := model e1 <== model e2.

Lemma prog_refl e : pleq e e. 
Proof. by []. Qed.

Lemma prog_trans e1 e2 e3 : pleq e1 e2 -> pleq e2 e3 -> pleq e1 e3.
Proof. by move=>H1 H2 p x; apply: H2; apply: H1. Qed.

Lemma prog_asym e1 e2 : pleq e1 e2 -> pleq e2 e1 -> e1 = e2.
Proof.
move: e1 e2=>[m1 N1][m2 N2]; rewrite /pleq /= => H1 H2.
by rewrite (poset_asym H1 H2) in N1 N2 *; congr Prog; apply: proof_irrelevance.
Qed.

Definition pbot_model : Pred (tree A) := Pred0.

Lemma pbot_spec : pbot_model <=p has_spec s.
Proof. by []. Qed.

Definition pbot := Prog pbot_spec.

Lemma prog_bot p : pleq pbot p.
Proof. by move=>t. Qed.

Definition progPosetMixin := 
  PosetMixin prog_bot prog_refl prog_asym prog_trans.
Canonical progPoset := Eval hnf in Poset _ progPosetMixin.

Definition psup_model (es : Pred (prog s)) := 
  [Pred t : tree A | exists e, t \In model e /\ e \In es].

Lemma psup_spec es : psup_model es <=p has_spec s.
Proof. by move=>p1; case=>[[m1] N][]; move/N. Qed.

Definition psup es := Prog (@psup_spec es).

(* sup is an upper bound *)
Lemma prog_supB es p : p \In es -> p <== psup es.
Proof. by move=>H1 [|v|B a p'|B1 B2 p1 pE p']; exists p. Qed.

(* sup is the least upper bound *)
Lemma prog_supL es p : (forall q, q \In es -> q <== p) -> psup es <== p.
Proof. by case: p=>[md] N H [|v|B a g p'|B1 B2 p1 pE p'] [p] []; move/H. Qed.

Definition progLatticeMixin := LatticeMixin prog_supB prog_supL.
Canonical progLattice := Lattice _ progLatticeMixin.

End Lattice.

(*********************)
(* Basic combinators *)
(*********************)

(* recursion *)

Section Fix. 
Variables (A : Type) (B : A -> Type) (s : forall x, spec (B x)).
Notation tp := (forall x, prog (s x)).
Notation lat := (dfunLattice (fun x => [lattice of prog (s x)])).
Variable (f : tp -> tp).

(* we take a fixpoint not of f, but of its monotone completion f' *)
(* should eventually prove that f' is monotone *)

(* XXX need to add idle prefix to ensure non-emptiness *)
Definition f' (e : lat) := 
  sup [Pred t : lat | exists e', e' <== e /\ t = f e'].

Definition ffix : tp := tarski_lfp f'. 

End Fix.


(* monadic unit *)

Section Return.
Variables (A : Type) (x : A).

Definition ret_s : spec A := 
  (fun gE lp => True, 
   fun v gE1 gE2 lp1 lp2 => 
     [/\ lp2 = lp1, v = x & 
         owns (mx lp2) -> gE2 = gE1]).

Definition ret_model : Pred (tree A) := Pred1 (Ret x).

Lemma ret_has_spec : ret_model <=p has_spec ret_s.
Proof. 
move=>p [<-] sh lp ep C _; split; apply: alw_retP=>//.
case=>sh' lp' ep' [/= <- H' _] v' [<-]; split=>//.
by move/H'.
Qed.

Definition ret := Prog ret_has_spec. 

End Return.

(* monadic bind *)

Section Bind.
Variables (A B : Type) (s1 : spec A) (s2 : A -> spec B).
Variables (e1 : prog s1) (e2 : forall x, prog (s2 x)).

Definition bind_s : spec B := 
  (fun gE1 lp1 => 
     s1.1 gE1 lp1 /\
     forall x2 sh2 lp2 ep2, 
       [sh2, lp2, ep2] \In coherent ->
       s1.2 x2 gE1 (gh ep2) lp1 lp2 -> 
       (s2 x2).1 (gh ep2) lp2,
   fun x2 gE gE2 lp lp2 => 
     exists x1, exists sh1, exists lp1, exists ep1, 
       [/\ [sh1, lp1, ep1] \In coherent, 
           s1.2 x1 gE (gh ep1) lp lp1 
         & (s2 x1).2 x2 (gh ep1) gE2 lp1 lp2]).

(* hmm, is the above enough?  i'm worried about being able to say that if
the local owns the lock, then the env ghost doesn't change. but maybe 
that should be specified by individual commands *)

Definition bind_model := 
  [Pred p | exists q, q \In model e1 /\ 
              p \In tcatp q (fun v => [PredU model (e2 v) & Pred1 Dummy])].

Lemma bind_has_spec : bind_model <=p has_spec bind_s.
Proof.
move=>p12 [p1][]; move/specE=>H1 H12 sh1 lp1 ep1 C1 /= [S1] T1.
split.
- apply: lsf_bnd H12 _; move/(_ _ _ _ C1 S1)/proj2: H1.
  apply: aft_imp=>[[sh2 lp2 ep2]] v2 C2; move/T1=>S2 p.
  case=>/=; last by case=><-; apply: lsf_dummyP.
  by case/specE/(_ _ _  _ C2 (S2 _ C2)). 
apply: aft_bnd H12 _; move/(_ _ _ _ C1 S1)/proj2: H1.
apply: aft_imp=>[[sh2 lp2 ep2]] v2 C2 S2; move/T1: (S2)=>T2 p.
case=>/=; last by case=><-; apply: alw_dummyP.
move/specE/(_ _ _ _ C2 (T2 _ C2))/proj2.
apply: aft_imp=>[[sh3 lp3 ep3]] v3 C3 S3.
by exists v2, sh2, lp2, ep2.
Qed.

Definition bind := Prog bind_has_spec.

End Bind.

Section Par.
Variables (A B : Type) (s1 : spec A) (s2 : spec B).
Variables (e1 : prog s1) (e2 : prog s2).

Definition par_s : spec (A * B) := 
  (fun gE lp => exists lp1, exists lp2, 
     [/\ lp = lp1 \+ lp2, s1.1 (gh lp2 \+ gE) lp1 & s2.1 (gh lp1 \+ gE) lp2], 
   fun x gE gE' lp lp' => forall lp1 lp2, 
     lp = lp1 \+ lp2 -> s1.1 (gh lp2 \+ gE) lp1 -> s2.1 (gh lp1 \+ gE) lp2 ->
     exists lp1', exists lp2',
       [/\ lp' = lp1' \+ lp2', 
           s1.2 x.1 (gh lp2 \+ gE) (gh lp2' \+ gE') lp1 lp1' 
         & s2.2 x.2 (gh lp1 \+ gE) (gh lp1' \+ gE') lp2 lp2']).
 
Definition par_model := 
  [Pred p12 | exists p1, exists p2, 
     [/\ p12 = Par p1 p2 Ret, p1 \In model e1 & p2 \In model e2]].

Lemma par_has_spec : par_model <=p has_spec par_s.
Proof.
move=>_ [p1][p2][->]; move/specE=>H1; move/specE=>H2.
move=>sh lp ep C /= [lp1][lp2][E] S1 S2; rewrite {lp}E in C *. 
have Ls: lock_safe [sh, lp1 \+ lp2, ep] (Par p1 p2 Ret).
  have C1 : [sh, lp1, lp2 \+ ep] \In coherent.
  - by apply/coh_shuffle.
  have C2 : [sh, lp2, lp1 \+ ep] \In coherent.
  - by apply/coh_shuffle; rewrite joinC.
  move/(_ _ _ _ C1 S1)/proj1/aft_lsf: H1=>Ls1.
  move/(_ _ _ _ C2 S2)/proj1/aft_lsf: H2=>Ls2.
  by apply/aft_lsf; apply: aft_imp (aft_par Ls1 Ls2).
split=>//; rewrite (st_feta (fun s' => _)).
rewrite (aftA Ls)=>mp1; rewrite (aftA Ls)=>mp2.
rewrite !(aftA Ls)=>E {Ls S1 S2} S1 S2; rewrite {}E in C *.
have C1 : [sh, mp1, mp2 \+ ep] \In coherent.
- by apply/coh_shuffle.
have C2 : [sh, mp2, mp1 \+ ep] \In coherent.
- by apply/coh_shuffle; rewrite joinC.
move/(_ _ _ _ C1 S1)/proj2: H1=>Ha1.
move/(_ _ _ _ C2 S2)/proj2: H2=>Ha2.
apply: {Ha1 Ha2} aft_imp (aft_par Ha1 Ha2).
move=>s v _ [sh'][lp1'][lp2'][ep'][->] P1 P2 /=.
by exists lp1', lp2'.
Qed.

Definition par := Prog par_has_spec.

End Par.

(* frame is a simplified case of Par, where the second program is Idle *)
(* we do it as a program command now, but it should *)
(* be moved into the verified predicate eventually *)

Section Frame.
Variables (A : Type) (s : spec A) (e : prog s) (h : procHeap).

Definition frame_s : spec A := 
  (fun gE lp => exists lp1, lp = lp1 \+ h /\ s.1 (gh h \+ gE) lp1,
   fun x gE gE' lp lp' => forall lp1, 
     lp = lp1 \+ h -> s.1 (gh h \+ gE) lp1 ->
     exists lp1', lp' = lp1' \+ h /\ 
                  s.2 x (gh h \+ gE) (gh h \+ gE') lp1 lp1').

Definition frame_model := model e.

Lemma frame_has_spec : frame_model <=p has_spec frame_s.
Proof.
move=>t; move/specE=>H1 sh lp ep C T1.
set P := after _ _ _; suff H: P.
- by split=>//; apply/aft_lsf; apply: aft_imp H.
rewrite {}/P. 
case: T1=>lp1 [E] H; rewrite {lp}E in C *.
have Ls : lock_safe [sh, lp1 \+ h, ep] t.
- move/coh_shuffle: C=>C.
  case: (H1 _ _ _ C H); rewrite joinC; move/alw_frame=>Ls _.
  by apply: alw_imp Ls.
(* simplify the goal to pull implications out *)
suff : after [sh, lp1 \+ h, ep] t
         (fun s' v' => forall lp0,
           lp1 \+ h = lp0 \+ h ->
           s.1 (gh h \+ gh ep) lp0 ->
           exists lp1' : procHeap,
             local s' = lp1' \+ h /\
             s.2 v' (gh h \+ gh ep) (gh h \+ gh (env s')) lp0 lp1').
- by apply: aft_imp; case. 
rewrite (aftA Ls)=>lp'. 
rewrite !(aftA Ls)=>E {H} H. 
rewrite E in C *; move/coh_shuffle: C=>C.
case: (H1 _ _ _ C H); rewrite joinC=>_.
move/alw_frame; apply: alw_imp.
case=>sh' lp1' ep' p' /= C' Sf [lp2][E1] L v T.
by exists lp2; rewrite !(joinC (gh h)); split=>//; apply: L. 
Qed.

End Frame.

(*****************************)
(* Singleton heap operations *)
(*****************************)

Section Read.
Variables (A : Type) (x : ptr).

Definition read_s : spec A := 
  (fun gE lp => exists h, exists v : A, 
     lp = [Mgh mx lp, gh lp, x :-> v :+ h],
   fun y gE1 gE2 lp1 lp2 => 
     forall h (v : A), 
       lp1 = [Mgh mx lp1, gh lp1, x :-> v :+ h] ->
       [/\ lp2 = lp1, y = v &
           owns (mx lp2) -> gE2 = gE1]).

Definition read_model : Pred (tree A) := Pred1 (@Read A x @ Ret).

Lemma read_has_spec : read_model <=p has_spec read_s.
Proof.
move=>_ [<-] sh [lx lg lh] ep C [h][v][E]; rewrite {lh}E in C *.
set P := after _ _ _; suff H: P.
- by split=>//; apply/aft_lsf; apply: aft_imp H.
apply: {P} alw_cons C _.
case=>sh1 _ ep1 [/= <- H1 C1]; split=>//.
- apply/sf_consP=>_; split=>//=.
  apply/asf_readP; first by apply: defUnl (cohD C1).
  by exists (sh1 :+ h), v; rewrite unCA.
move=>w _ [/= ->].
rewrite -!unA unCA lookPtUn; last first.
- by move: (cohD C1); defcheck.
move/dyn_inj=>->{w}.
exists sh1, [Mgh lx, lg, x :-> v :+ h], id_annot; split=>/=.
- by rewrite -!unA unCA.
- by exists A, (Read x).
- by apply/csP1.
apply: alw_retP. 
- apply: coh_sub (C1); split=>//=.
  by move: (cohD C1); defcheck.
case=>sh' _ ep' [/= <- H' C'] _ [<-] j w []. 
have D: def (x :-> v :+ h) by move: (cohD C1); defcheck.
(* XXX probably can use look injectivity instead of cancel *)
case/(cancel D)=>-> _ ->; split=>//=.
by move=>O; rewrite -H' // H1.
Qed.

Definition read := Prog read_has_spec.

End Read.

Section Write.
Variables (A : Type) (x : ptr) (v : A).

Definition write_s : spec unit := 
  (fun gE lp => exists h, exists B, exists w : B, 
     lp = [Mgh mx lp, gh lp, x :-> w :+ h],
   fun _ gE1 gE2 lp1 lp2 => 
     forall h B (w : B), 
       lp1 = [Mgh mx lp1, gh lp1, x :-> w :+ h] -> 
       lp2 = [Mgh mx lp1, gh lp1, x :-> v :+ h] /\
       (owns (mx lp2) -> gE2 = gE1)).

Definition write_model : Pred (tree unit) := Pred1 (Write x v @ Ret).

Lemma write_has_spec : write_model <=p has_spec write_s.
Proof.
move=>_ [<-] sh [lx lg lh] ep C [h][B][w] /= E; rewrite {lh}E in C *.
set P := after _ _ _; suff H: P.
- by split=>//; apply/aft_lsf; apply: aft_imp H.
apply: {P} alw_cons C _.
case=>sh1 _ ep1 [/= <- H1 C1]; split=>//.
- apply/sf_consP=>_; split=>//=.
  apply/asf_writeP; first by apply: defUnl (cohD C1).
  by exists (sh1 :+ h), B, w; rewrite unCA.
move=>[] _ [/= ->].
exists sh1, [Mgh lx, lg, x :-> v :+ h], id_annot; split=>/=.
- by rewrite -!unA -!(unCA (x :-> _)) updPtUn.
- by exists unit, (Write x v).
- by apply/csP1.
apply: alw_retP. 
- apply: coh_sub (C1); split=>//=.
  by move: (cohD C1); defcheck.
case=>sh' _ ep' [/= <- H' C'][] _ j C u [E]. 
have D: def (x :-> w :+ h) by move: (cohD C1); defcheck.
(* XXX probably can use look injectivity instead of cancel *)
move/(cancelT D): (E)=>F; rewrite -{C}F in u E *.
case/(cancel D): E=>_ _ ->; split=>//=.
by move=>O; rewrite -H' // -H1.
Qed.

Definition write := Prog write_has_spec.

End Write.

Section Allocation.
Variables (A : Type) (v : A).

Definition alloc_s : spec ptr := 
  (fun gE lp => True,
   fun x gE1 gE2 lp1 lp2 => 
     lp2 = [Mgh mx lp1, gh lp1, x :-> v :+ lp1] /\
     (owns (mx lp2) -> gE2 = gE1)).

Definition alloc_model : Pred (tree ptr) := Pred1 (Alloc v @ Ret).

Lemma alloc_has_spec : alloc_model <=p has_spec alloc_s.
Proof.
move=>_ [<-] sh lp ep C []; set P := after _ _ _; suff H: P.
- by split=>//; apply/aft_lsf; apply: aft_imp H.
apply: {P} alw_cons C _.
case=>sh1 _ ep1 [/= <- H1 C1]; split=>//; first by apply/sf_consP.
move=>x m [->] D1 N2.
exists sh1, [Mgh mx lp, gh lp, lp :+ x :-> v], id_annot=>/=; split.
- by rewrite !unA -(unAC _ (x :-> v)) -(unC (x :-> v)).
- by exists ptr, (Alloc v).
- by apply/csP1.
apply: alw_retP; last first.
- case=>sh' _ ep' [/= <- H' C'][_][<-]; rewrite unC; split=>//=.
  by move=>O; rewrite -H' // -H1.
apply: coh_sub (C1).
by rewrite -!unA !(unCA _ (x :-> v)) defPtUn N2 unA (cohD C1) D1.
Qed.

Definition alloc := Prog alloc_has_spec.

End Allocation.

Section Deallocation.
Variable x : ptr.

Definition dealloc_s : spec unit := 
  (fun gE lp => 
     exists h, exists A, exists v : A, 
       lp = [Mgh mx lp, gh lp, x :-> v :+ h],
   fun _ gE1 gE2 lp1 lp2 => 
     forall h A (v : A), 
       lp1 = [Mgh mx lp1, gh lp1, x :-> v :+ h] -> 
       lp2 = [Mgh mx lp1, gh lp1, h] /\
       (owns (mx lp2) -> gE2 = gE1)).

Definition dealloc_model : Pred (tree unit) := Pred1 (Dealloc x @ Ret).

Lemma dealloc_has_spec : dealloc_model <=p has_spec dealloc_s.
Proof.
move=>_ [<-] sh [lx lg lh] ep C [h][A][v] /= E; rewrite {lh}E in C *.
set P := after _ _ _; suff H: P.
- by split=>//; apply/aft_lsf; apply: aft_imp H.
apply: {P} alw_cons C _.
case=>sh1 _ ep1 [/= <- H1 C1]; split=>//.
- apply/sf_consP=>_; split=>//=.
  apply/asf_deallocP; first by apply: defUnl (cohD C1).
  by exists (sh1 :+ h), A, v; rewrite unCA.
move=>[] _ [/= ->].
exists sh1, [Mgh lx, lg, h], id_annot; split=>/=.
- rewrite -!unA unCA freePtUn //.
  by move: (cohD C1); rewrite /= -!unA unCA.
- by exists unit, (Dealloc x).
- by apply/csP1.
apply: alw_retP. 
- apply: coh_sub (C1); split=>//=.
  by move: (cohD C1); defcheck.
case=>sh' _ ep' [/= <- H' C'][] _ j B w [].
have D: def (x :-> v :+ h) by move: (cohD C1); defcheck.
(* XXX probably can use look injectivity instead of cancel *)
move=>E; move/(cancelT D): (E)=>F; rewrite -{B}F in w E *.
case/(cancel D): E=>_ _ ->; split=>//=.
by move=>O; rewrite -H' // -H1.
Qed.

Definition dealloc := Prog dealloc_has_spec.

End Deallocation. 

(* I make CAS only possible on the shared state. *)
(* There's little point in doing it elsewhere, since *)
(* the semantics enforces protocol safety only on the shared state *)

Section CAS.

Definition cas_s : spec bool := 
  (fun gE lp => ownsno (mx lp) : Prop,
   fun y gE1 gE2 lp1 lp2 => 
     if y then 
       exists inv, 
         lp2 = [mgh true, gh lp1, inv :+ lp1] /\ 
         Inv (gh lp2 \+ gE2) inv
     else lp2 = lp1).

Definition cas_model : Pred (tree bool) := Pred1 (CAS lk Unlocked Locked @ Ret).

Lemma cas_has_spec : cas_model <=p has_spec cas_s.
Proof.
move=>_ [<-] sh lp ep C _; set P := after _ _ _; suff H: P.
- by split=>//; apply/aft_lsf; apply: aft_imp H.
apply: {P} alw_cons C _.
case=>sh1 _ ep1 [/= <- _ C1]; split=>//.
- apply/sf_consP=>_; split=>//=.
  apply/asf_casP; first by apply: defUnl (cohD C1).
  exists (lp :+ payload sh1), (if lockd sh1 then Locked else Unlocked).
  by rewrite (shr_eta sh1) -unA (unC lp). 
move=>v _ [w][E1] E2 ->; move: (cohD C1)=>/= D.
move: E1 E2; rewrite (shr_eta sh1) -!unA in D *.
rewrite lookPtUn // {D}; move/dyn_inj=><-{w} ->{v}.
case L1: (lockd sh1).
- exists sh1, lp, id_annot; split=>//=.
  - by rewrite (shr_eta sh1) L1 -!unA.
  - by exists bool, (CAS lk Unlocked Locked).
  - by rewrite /case_step L1.
  apply: alw_retP=>//.
  by case=>sh' _ ep' [/= <- H' C'] _ [<-].
exists (grant sh1), [mgh true, gh lp, lp :+ payload sh1], id_annot; split=>/=.
- by rewrite -!unA updPtUn grantE (unCA lp).
- by exists bool, (CAS lk Unlocked Locked).
- rewrite /case_step  L1 lockdG. 
  case: C1=>_; case/andP=>_ -> _; rewrite L1 /= => I; split=>//.
  by apply: I.
apply: alw_retP.
- apply: cohGlp.
  rewrite (cohD C1); move/cohV/andP/proj2: (C1)=>->.
  by move: (coh_lkd_owns C1); rewrite L1; case/andP=>_ ->.
case=>sh' _ ep' [/= <- H' C'] _ [<-] /=; rewrite unC.
exists (payload sh1); split=>//; rewrite -H' //.
by case: C1=>_ _ _ /=; rewrite L1; apply. 
Qed.

Definition cas := Prog cas_has_spec.

End CAS.

(* we need to do unlock primitively, because it messes with shared state *)
(* avoid using :- in the specs, to simplify the client reasoning *)

Section Unlock.
(* G : U -> Pred U, i.e., a function of the old ghost *)
(* G : U -> U, i.e., a function of the old ghost *)
(* Variables (G : U -> U). *)
Variables (G : annot U).
(* Variables (G : Pred U) (Gfunc : functional_pred G). *)
Notation Inv_prec := (@mod_inv_prec R).
(* Variables (Inv_prec : forall g, precise (Inv (R:=R) g)). *)

Definition unlock_s : spec unit :=
  (fun gE lp => 
     exists h, exists inv, exists gL,
       [/\ lp = [mgh true, gh lp, inv :+ h], 
           assert G (gh lp) = Some gL,
           valid (gL \+ gE) &
           Inv (gL \+ gE) inv],
   fun _ gE1 gE2 lp1 lp2 =>
     forall h inv gL, 
       lp1 = [mgh true, gh lp1, inv :+ h] ->
       assert G (gh lp1) = Some gL ->
       valid (gL \+ gE1) ->
       Inv (gL \+ gE1) inv ->
       lp2 = [mgh false, gL, h]).

Definition unlock_model : Pred (tree unit) := Pred1 (Cons (Write lk Unlocked) G Ret).

Lemma unlock_has_spec : unlock_model <=p has_spec unlock_s.
Proof.
move=>_ [<-] sh lp ep /= C [h][inv][gL][E] Asrt V I; rewrite {}E /= in C *.
set P := after _ _ _; suff H: P.
- by split=>//; apply/aft_lsf; apply: aft_imp H.
apply: {P} alw_cons C _; case=>sh1 _ ep1 [/= <- O C1].
rewrite {O ep}(O (erefl _)) in I V *.
move: (coh_owns_lkd C1)=>/= L1. 
split=>//.
- apply/sf_consP=>_; split=>//=.
  apply/asf_writeP; first by apply: defUnl (cohD C1).
  exists (payload sh1 :+ inv :+ h), lockb, Locked.
  by rewrite (shr_eta sh1) L1 !unA.
case=>_ ->.
exists (receive sh1 inv), [mgh false, gL, h], G.
split=>/=.
- rewrite (shr_eta sh1) L1 payloadE // unh0 -!unA.
  rewrite updPtUn unA receiveE //.
  by move: (cohD C1); defcheck.
- by exists unit, (Write lk Unlocked).
- rewrite /case_step L1 lockdR; split=>//.
  exists inv, gL; split=>//.
  - by rewrite unC unhSh //; move: (cohD C1); defcheck.
  - by apply: subheapUnl; move: (cohD C1); defcheck.
apply: alw_retP.
- split=>/=; last first.
  - by rewrite payloadR //; move: (cohD C1); defcheck.
  - by case: (mx ep1) (coh_ownsL C1).
  - by rewrite /valid /= V; case/andP: (cohV C1); case: (mx ep1).
  move: (cohD C1)=>/= D.
  rewrite receiveE; last by move: D; defcheck.
  rewrite (shr_eta sh1) payloadE L1 // unh0 -!unA in D.
  by rewrite -!unA !defPtUn in D *.
case=>sh' _ ep' [/= <- _ C'][] _ h' inv' gL' [] E Asrt2 V2 I2.
have D: def (inv :+ h) by move: (cohD C1); defcheck.
have Sh1 : subheap inv (inv :+ h) by apply:(subheapUnl D).
have Sh2 : subheap inv' (inv :+ h) by rewrite E in D *; apply:(subheapUnl D).
rewrite Asrt in Asrt2; case: Asrt2=>Asrt2; rewrite -{gL'}Asrt2 in V2 I2 *.
rewrite -(Inv_prec Sh1 I Sh2 I2) in E.
by rewrite -(unhKl D E).
Qed.

Definition unlock := Prog unlock_has_spec.

End Unlock.

(***************)
(* Consequence *)
(***************)

Definition verify A lp gE (s : spec A) (r : cont A) := 
  forall sh mE hE, [sh, lp, [Mgh mE, gE, hE]] \In coherent -> 
  s.1 gE lp /\ 
  forall x sh' lp' mE' gE' hE', 
    [sh', lp', [Mgh mE', gE', hE']] \In coherent -> 
    s.2 x gE gE' lp lp' -> r x gE' lp'.


Definition conseq A (s : spec A) (p : pre) (q : post A) := 
  forall lp gE, p gE lp -> 
    verify lp gE s (fun x' gE' lp' => q x' gE gE' lp lp').

Local Notation conseq1 := 
  (fun A (s1 s2 : spec A) => 
     conseq s1 (let 'pair x _ := s2 in x) 
               (let 'pair _ x := s2 in x)).

Lemma conseq_refl A (s : spec A) : conseq1 A s s. 
Proof. by case: s. Qed.

Hint Resolve conseq_refl. 

Section Consequence.
Variables (A : Type) (s1 s2 : spec A).
Variables (e : prog s1) (pf : conseq1 A s1 s2). 

Definition do_model := model e.

Lemma do_has_spec : do_model <=p has_spec s2.
Proof.
move=>t; move/specE=>H1 sh lp ep C T1.
set P := after _ _ _; suff H: P.
- by split=>//; apply/aft_lsf; apply: aft_imp H.
rewrite {}/P; case: s2 T1 pf=>p2 q2 /= T1. 
case/(_ _ _ T1 _ _ _ C)=>S1 S2.
move/(_ _ _ _ C S1)/proj2: H1.
apply: aft_imp; case=>sh' lp' ep' v' C'.
by apply: S2 C'.
Qed.

Definition do' := Prog do_has_spec. 

End Consequence.

Notation "'Do' e" := (do' e _) (at level 80).

(**********************************)
(* Conditionals for various types *)
(**********************************)

Section CondBool.
Variables (A : Type) (b : bool) (s1 s2 : spec A). 

Program
Definition If (e1 : prog s1) (e2 : prog s2) : prog (if b then s1 else s2) :=
  match b with true => Do e1 | false => Do e2 end.
End CondBool.

Section CondDecide.
Variable (A : Type) (p1 p2 : Prop) (b : {p1} + {p2}) 
         (s1 : p1 -> spec A) (s2 : p2 -> spec A).

Program
Definition Match_dec (e1 : forall x, prog (s1 x))
                     (e2 : forall x, prog (s2 x)) : 
             prog (match b with left x => s1 x | right x => s2 x end) :=
  match b with left x => Do (e1 x) | right x => Do (e2 x) end.
End CondDecide. 

Section CondList.
Variables (A B : Type) (xs : seq A) (s1 : spec B) (s2 : A -> seq A -> spec B).

Program
Definition Match_list xs (e1 : prog s1) (e2 : forall x' xs', prog (s2 x' xs'))
  : prog (match xs with nil => s1 | cons x' xs' => s2 x' xs' end) :=
  match xs with nil => Do e1 | cons x' xs' => Do (e2 x' xs') end.
End CondList.

(* add more conditionals here *)

(********************************************)
(* Derived logical rules for easier proving *)
(********************************************)

Lemma val_do A (s : spec A) lp gE (r : cont A) :
         s.1 gE lp -> 
         (forall x lp' gE', 
            s.2 x gE gE' lp lp' -> 
            valid (gh lp' \+ gE') ->
            r x gE' lp') ->
         verify lp gE s r.
Proof.
move=>H1 H2 sh C; split=>// x sh' lp' mE' gE' hE' C' S'; apply: H2 S' _.
case: C'=>/= D; case/andP=>/= V1 V2 _ _. 
by rewrite /valid /=; move: D; defcheck.
Qed.

Lemma bnd_do A B (s : spec A) s2 lp gE (r : cont B) : 
        s.1 gE lp -> 
        (forall x lp' gE', 
           s.2 x gE gE' lp lp' -> 
           verify lp' gE' (s2 x) r) -> 
        verify lp gE (bind_s s s2) r.
Proof.
move=>S T C; split; first by split=>// x2 sh2 lp2 ep2 C2; case/T/(_ _ _ _ C2). 
move=>x sh' lp' mE' gE' hE' C' [x1][sh1][lp1][ep1][C1] S2 T2.
by move/T/(_ _ _ _ C1)/proj2/(_ _ sh' lp' mE' gE' hE'): S2; apply.
Qed.

(* in par we need to supply intermediate annotations r1 and r2 *)
(* hopefully, these will usually be the default ones, e.g. *)
(* from the postcondition of e1 and e2; so we won't have to actually *)
(* write them explicitly *)

Lemma par_do A B (s1 : spec A) (s2 : spec B) lp1 lp2 gE
                 (r1 : cont A) (r2 : cont B) (r : cont (A * B)) :
        verify lp1 (gh lp2 \+ gE) s1 r1 ->
        verify lp2 (gh lp1 \+ gE) s2 r2 ->
        (forall x1 x2 lp1 lp2 gE, 
           valid (gh lp1 \+ gh lp2 \+ gE) ->
           r1 x1 (gh lp2 \+ gE) lp1 -> r2 x2 (gh lp1 \+ gE) lp2 ->
           r (x1, x2) gE (lp1 \+ lp2)) ->
        verify (lp1 \+ lp2) gE (par_s s1 s2) r.
Proof.
move=>H1 H2 H sh mE hE C. 
have C1 : [sh, lp1, lp2 \+ [Mgh mE, gE, hE]] \In coherent.
- by apply/coh_shuffle.
have C2 : [sh, lp2, lp1 \+ [Mgh mE, gE, hE]] \In coherent.
- by apply/coh_shuffle; rewrite joinC.
have S1 : s1.1 (gh lp2 \+ gE) lp1 by case/H1: C1.
have S2 : s2.1 (gh lp1 \+ gE) lp2 by case/H2: C2.
split; first by exists lp1; exists lp2.
case=>x1 x2 sh' lp' mE' gE' hE' C' T.
case/(_ lp1 lp2 (erefl _) S1 S2): T C'=>lp1' [lp2'][->] T1 T2 C'. 
apply: H.
- case: C'=>/= D'; case/andP=>/= V1 V2 _ _.
  by rewrite /valid /=; move: D'; defcheck.
- move/(_ _ _ _ C1)/proj2/(_ _ sh' lp1'): H1.
  by move/coh_shuffle: C'=>C'; move/(_ _ _ _ _ C'); apply.
move/(_ _ _ _ C2)/proj2/(_ _ sh'): H2; move: C'; rewrite joinC.
by move/coh_shuffle=>C'; move/(_ _ _ _ _ _ C'); apply.
Qed.

Lemma val_ret A x lp gE (r : cont A) : 
       (forall gE', 
          (owns (mx lp) -> gE' = gE) ->
          valid (gh lp \+ gE') ->
          r x gE' lp) -> 
       verify lp gE (ret_s x) r. 
Proof.
move=>T; apply: val_do=>// y lp' ep' S' V'.
by case: S'=>E -> H'; rewrite {lp'}E in V' H' *; apply: T H' V'.
Qed.

Lemma bnd_ret A B (x : A) lp gE s (r : cont B) : 
        (forall gE', 
           (owns (mx lp) -> gE' = gE) ->
           verify lp gE' (s x) r) -> 
        verify lp gE (bind_s (ret_s x) s) r.
Proof. 
by move=>T; apply: bnd_do=>// y lp' ep' [->] -> H'; apply: T H'.
Qed.

Lemma val_read A x v lx lg gE h (r : cont A) : 
        (forall gE', 
           (owns lx -> gE' = gE) ->
           valid (lg \+ gE') ->
           r v gE' [Mgh lx, lg, x :-> v :+ h]) ->
        verify [Mgh lx, lg, x :-> v :+ h] gE (read_s A x) r.
Proof.
move=>T; apply: val_do=>/=; first by exists h, v.
by move=>y lp' ep'; case/(_ _ _ (erefl _))=>->->; apply: T.
Qed.
 
Lemma bnd_read A B x v lx lg gE h s (r : cont B) : 
        (forall gE', 
           (owns lx -> gE' = gE) -> 
           verify [Mgh lx, lg, x :-> v :+ h] gE' (s v) r) -> 
        verify [Mgh lx, lg, x :-> v :+ h] gE (bind_s (read_s A x) s) r.
Proof.
move=>T; apply: bnd_do=>/=; first by exists h, v.
by move=>y lp' ep'; case/(_ _ _ (erefl _))=>->->; apply: T.
Qed.

Lemma val_write A B x (v : A) (w : B) lx lg gE h (r : cont unit) : 
        (forall gE', 
           (owns lx -> gE' = gE) ->
           valid (lg \+ gE') ->
           r tt gE' [Mgh lx, lg, x :-> v :+ h]) ->
        verify [Mgh lx, lg, x :-> w :+ h] gE (write_s x v) r.
Proof.
move=>T; apply: val_do=>/=; first by exists h, B, w.
by move=>[] lp' ep'; case/(_ _ _ _ (erefl _))=>->; apply: T.
Qed.
 
Lemma bnd_write A B C x (v : A) (w : B) lx lg gE h s (r : cont C) : 
        (forall gE', 
           (owns lx -> gE' = gE) -> 
           verify [Mgh lx, lg, x :-> v :+ h] gE' (s tt) r) -> 
        verify [Mgh lx, lg, x :-> w :+ h] gE (bind_s (write_s x v) s) r.
Proof.
move=>T; apply: bnd_do=>/=; first by exists h, B, w.
by move=>[] lp' ep'; case/(_ _ _ _ (erefl _))=>->; apply: T.
Qed.

(* Q: do we need here that y is not equal to anything *)
(* in the shared resource? It seems not. In bnd_alloc *)
(* the inner verify takes care of that. In val_alloc  *)
(* r is independent of sh, so it doesn't matter.      *)

Lemma val_alloc A (v : A) lx lg h gE (r : cont ptr) : 
        (forall y gE', 
           (owns lx -> gE' = gE) ->
           valid (lg \+ gE') ->
           r y gE' [Mgh lx, lg, y :-> v :+ h]) -> 
        verify [Mgh lx, lg, h] gE (alloc_s v) r.
Proof. by move=>T; apply: val_do=>//= y lp' ep' [->]; apply: T. Qed.

Lemma bnd_alloc A (v : A) C lx lg h gE s (r : cont C) : 
        (forall y gE', 
           (owns lx -> gE' = gE) ->
           verify [Mgh lx, lg, y :-> v :+ h] gE' (s y) r) ->
        verify [Mgh lx, lg, h] gE (bind_s (alloc_s v) s) r.
Proof. by move=>T; apply: bnd_do=>//= y _ ep' [->]; apply: T. Qed.

Lemma val_dealloc A x (v : A) lx lg h gE (r : cont unit) : 
        (forall gE',
           (owns lx -> gE' = gE) ->
           valid (lg \+ gE') ->
           r tt gE' [Mgh lx, lg, h]) -> 
        verify [Mgh lx, lg, x :-> v :+ h] gE (dealloc_s x) r.
Proof.
move=>T; apply: val_do=>/=; first by exists h, A, v.
by move=>[] lp' ep'; case/(_ _ _ _ (erefl _))=>->; apply: T.
Qed.

Lemma bnd_dealloc A B x (v : B) lx lg h gE s (r : cont A) : 
        (forall gE',
           (owns lx -> gE' = gE) ->
           verify [Mgh lx, lg, h] gE' (s tt) r) ->
        verify [Mgh lx, lg, x :-> v :+ h] gE (bind_s (dealloc_s x) s) r.
Proof.
move=>T; apply: bnd_do=>/=; first by exists h, B, v.
by move=>[] lp' ep'; case/(_ _ _ _ (erefl _))=>->; apply: T.
Qed.

Lemma val_cas lg lh ep (r : cont bool) : 
        (forall gE' inv,
           Inv (lg \+ gE') inv ->
           valid (lg \+ gE') ->
           r true gE' [mgh true, lg, inv :+ lh]) ->
        (forall gE', 
           valid (lg \+ gE') ->
           r false gE' [mgh false, lg, lh]) ->
        verify [mgh false, lg, lh] ep cas_s r.
Proof.
move=>T1 T2; apply: val_do=>//= x lp' ep'.
case: x; last by move=>->; apply: T2.
by case=>inv [->]; apply: T1.
Qed.

Lemma bnd_cas A lg lh gE s (r : cont A) : 
        (forall gE' inv,
           Inv (lg \+ gE') inv ->
           verify [mgh true, lg, inv :+ lh] gE' (s true) r) ->
        (forall gE', 
           verify [mgh false, lg, lh] gE' (s false) r) ->
        verify [mgh false, lg, lh] gE (bind_s cas_s s) r.
Proof.
move=>T1 T2; apply: bnd_do=>//= x lp' ep'.
case: x; last by move=>->; apply: T2.
by case=>inv [->]; apply: T1.
Qed.

Lemma val_unlock gL h inv gE lg (r : cont unit) G : 
        assert G lg = Some gL ->
        valid (gL \+ gE) -> Inv (gL \+ gE) inv ->
        (forall gE', 
           valid (gL \+ gE') ->
           r tt gE' [mgh false, gL, h]) ->
        verify [mgh true, lg, inv :+ h] gE (unlock_s G) r.
Proof.
move=>/= Asrt V I T; apply: val_do=>/=; first by exists h, inv, gL.
by move=>[] lp' ep'; move/(_ _ _ _ (erefl _) Asrt V I)=>->//; apply: T.
Qed.

Lemma bnd_unlock A gL h inv gE lg s (r : cont A) G : 
        assert G lg = Some gL ->
        valid (gL \+ gE) -> Inv (gL \+ gE) inv ->
        (forall gE', 
           verify [mgh false, gL, h] gE' (s tt) r) ->
        verify [mgh true, lg, inv :+ h] gE (bind_s (unlock_s G) s) r.
Proof.
move=>/=Asrt V I T; apply: bnd_do=>/=; first by exists h, inv, gL.
by move=>[] lp' ep'; move/(_ _ _ _ (erefl _) Asrt V I)=>->. 
Qed.

(*********************************************************)
(* rules for pulling out the spec vars from binary posts *)
(*********************************************************)

Lemma vrf_mono A lp ep s (r1 r2 : cont A) : 
        r1 <== r2 -> verify lp ep s r1 -> verify lp ep s r2.
Proof.
move=>T H1 sh mE hE C; case: (H1 sh _ _ C)=>S1 T1.
by split=>// x sh' lp' mE' gE' hE' C'; move/(T1 _ _ _ _ _ _ C'); apply: T.
Qed.

Lemma vrfV A lp gE s (r : cont A) : 
        (valid (gh lp \+ gE) -> verify lp gE s r) ->
        verify lp gE s r.
Proof. by move=>H sh mE hE C; apply: H (C); move: (cohVld C); case/and3P. Qed.
      
Section Ghosts.
Variables (A : Type) (s : spec A) (p : pre).

Lemma ghA (B1 B2 : Type) (q : B1 -> B2 -> post A) :
       conseq s p (fun y gE1 gE2 lp1 lp2 => 
                     forall x1 x2, q x1 x2 y gE1 gE2 lp1 lp2) <->
       conseq s p (fun y gE1 gE2 lp1 lp2 => 
                     forall x, q x.1 x.2 y gE1 gE2 lp1 lp2).
Proof.
split=>H1 lp1 ep1 H2; move: (H1 lp1 ep1 H2);
apply: vrf_mono=>x' lp' ep' H; first by case; apply: H.
by move=>x1 x2; apply: (H (x1, x2)).
Qed.
   
Lemma ghAI B (q1 q2 : B -> pre) (r : B -> post A) :
        conseq s p (fun y gE1 gE2 lp1 lp2 => 
                      forall x, q1 x gE1 lp1 -> q2 x gE1 lp1 -> 
                                r x y gE1 gE2 lp1 lp2) <->
        conseq s p (fun y gE1 gE2 lp1 lp2 => 
                      forall x, q1 x gE1 lp1 /\ q2 x gE1 lp1 -> 
                                r x y gE1 gE2 lp1 lp2).
Proof.
split=>H1 lp1 ep1 H2; move: (H1 lp1 ep1 H2);
apply: vrf_mono=>x' lp' ep' H x; first by case; apply: H.
by move=>Q1 Q2; apply: H.
Qed.

Lemma ghE (B : Type) (q : B -> pre) (r : B -> post A) : 
        (forall lp1 gE1, p gE1 lp1 -> exists x, q x gE1 lp1) ->
        (forall lp1 gE1 x, 
           q x gE1 lp1 -> p gE1 lp1 -> valid (gh lp1 \+ gE1) ->
           verify lp1 gE1 s (fun y gE2 lp2 => r x y gE1 gE2 lp1 lp2)) ->
        conseq s p (fun y gE1 gE2 lp1 lp2 => 
                      forall x, q x gE1 lp1 -> r x y gE1 gE2 lp1 lp2).
Proof.
move=>H1 T1 lp1 ep1 S1 sh mE hE C; split.
- case: (H1 _ _ S1)=>x Q1; move: (T1 _ _ _ Q1 S1).
  by case/and3P: (cohVld C)=>_ C' _; case/(_ C' _ _ _ C).
move=>y' sh' lp' mE' gE' hE' C' T' x Q1'.
case/and3P: (cohVld C)=>_ V' _.
by move/proj2/(_ _ _ _ _ _ _ C' T'): (T1 _ _ _ Q1' S1 V' _ _ _ C).
Qed.

End Ghosts.

Definition ghs := (ghA, ghAI).

End CSL_Logic.

Implicit Arguments prog [R].
Implicit Arguments spec [R].
Implicit Arguments ret [R A].
Implicit Arguments read [R].
Implicit Arguments write [R A].
Implicit Arguments alloc [R A].
Implicit Arguments dealloc [R].
Implicit Arguments cas [R].
Implicit Arguments unlock [R].

Implicit Arguments par_do [R A B s1 s2 lp1 lp2 gE r].

(* some notation *)

Notation "'Do' e" := (do' e _) (at level 80).
Notation "x '<--' c1 ';' c2" := (bind c1 (fun x => c2)) 
  (at level 78, right associativity) : proc_scope.
Notation "c1 ';;' c2" := (bind c1 (fun _ => c2)) 
  (at level 78, right associativity) : proc_scope.
Notation "'!' x" := (read _ x) (at level 50) : proc_scope.
Notation "e1 '::=' e2" := (write e1 e2) (at level 60) : proc_scope.
