Load loadpath.
Require Import ssreflect ssrbool ssrnat eqtype ssrfun seq rels ordtype.
Require Import prelude domains heaps pcm.
Set Implicit Arguments.
Unset Strict Implicit.
Import Prenex Implicits.

(******************)
(* heaps are pcms *)
(******************)

(* which means from now on, we can combine *)
(* local ghosts with local heaps, and env ghosts with env heaps. *)

Definition heapPCMMixin := PCMMixin unC unA un0h defUnl.
Canonical heapPCM := Eval hnf in PCM heap heapPCMMixin.

(* often we need to rewrite heap \+ into heap :+ *)
(* so that we can apply the lemmas from the heap library *)
(* this can be achieved by rewrite /PCM.join at specific places *)
(* but this will also unfold the other instances *)
(* So, we build a specific rule to do this *)

Lemma heap_join h1 h2 : h1 \+ h2 = h1 :+ h2.
Proof. by []. Qed.

(*****************************************)
(* Resource algebra for mutual exclusion *)
(*****************************************)

Module Mutex.

Inductive lockb := Locked | Unlocked.

Definition lockb_eq u v :=
  match u, v with
    Locked, Locked => true
  | Unlocked, Unlocked => true
  | _, _ => false
  end.  

Lemma lockb_eqP : Equality.axiom lockb_eq.
Proof. by do 2!case; constructor. Qed.

Definition lockb_eqMixin := EqMixin lockb_eqP.
Canonical lockb_eqType := Eval hnf in EqType lockb lockb_eqMixin.

(* Now the datatype for the contribution of both processes *)
(* with the possibility of contribution being undefined *)

Structure mutex := Invalid | ProcHasLock | ProcHasNoLock.

Definition mutex_eq g1 g2 := 
  match g1, g2 with
    Invalid, Invalid => true
  | ProcHasLock, ProcHasLock => true
  | ProcHasNoLock, ProcHasNoLock => true
  | _, _ => false
  end.

Lemma mutex_eqP : Equality.axiom mutex_eq.
Proof. by do 2!case; constructor. Qed.

Definition mutex_eqMixin := EqMixin mutex_eqP.
Canonical mutex_eqType := Eval hnf in EqType mutex mutex_eqMixin.

Definition valid g := 
  if g is Invalid then false else true.

Definition join g1 g2 := 
  match g1, g2 with
    ProcHasLock, ProcHasNoLock => ProcHasLock
  | ProcHasNoLock, ProcHasLock => ProcHasLock
  | ProcHasNoLock, ProcHasNoLock => ProcHasNoLock
  | _, _ => Invalid
  end.

Definition unit := ProcHasNoLock.

Notation "g1 \+ g2" := (join g1 g2) (at level 43, left associativity).

Lemma joinC g1 g2 : g1 \+ g2 = g2 \+ g1.
Proof. by case: g1; case: g2. Qed.

Lemma joinA g1 g2 g3 : g1 \+ (g2 \+ g3) = g1 \+ g2 \+ g3.
Proof. by case: g1; case: g2; case: g3. Qed.

Lemma joinKr g g1 g2 : valid (g \+ g1) -> g \+ g1 = g \+ g2 -> g1 = g2.
Proof. by case: g1; case: g2; case: g. Qed.

Lemma validL g1 g2 : valid (g1 \+ g2) -> valid g1.
Proof. by case: g1; case: g2. Qed.

Lemma unitL g : unit \+ g = g.
Proof. by case: g. Qed.

Module Exports.
Notation lockb := Mutex.lockb.
Notation mutex := Mutex.mutex.

(* somewhat annoyingly, some lemmas below require *)
(* exporting the following as well *)
Notation Locked := Mutex.Locked.
Notation Unlocked := Mutex.Unlocked.

Canonical Structure lockb_eqType.
Canonical Structure mutex_eqType.

Definition mutexPCMMixin := PCMMixin joinC joinA unitL validL.
Canonical mutexPCM := Eval hnf in PCM mutex mutexPCMMixin.

End Exports.

End Mutex.

Export Mutex.Exports.

(* There is the obvious isomorphism between mutex and option bool *)
(* We introduce some notation to simplify the treatment of *)
(* non-None cases *)

Definition bool2mutex (b : bool) := nosimpl
  if b then Mutex.ProcHasLock else Mutex.ProcHasNoLock.

Notation "[mx  b ]" := (bool2mutex b) (at level 0).

Notation owns mx := (mx == [mx true]).
Notation ownsno mx := (mx == [mx false]).

Lemma ownsE b : owns [mx b] = b.
Proof. by case: b. Qed.

Lemma mxE b1 b2 : [mx b1] == [mx b2] = (b1 == b2).
Proof. by case: b1; case: b2. Qed.

Lemma own_join_imp (mx1 mx2 : mutex) : 
        owns (mx1 \+ mx2) ==> owns mx1 || owns mx2.
Proof. by case: mx1; case: mx2. Qed.

Lemma own_joinE (mx1 mx2 : mutex) : 
        valid (mx1 \+ mx2) -> 
        owns (mx1 \+ mx2) = owns mx1 || owns mx2.
Proof. by case: mx1; case: mx2. Qed.

Lemma owno_join (mx1 mx2 : mutex) : 
        ownsno (mx1 \+ mx2) = ownsno mx1 && ownsno mx2.
Proof. by case: mx1; case: mx2. Qed.

(* The important law is the one for commuting lock *)
(* and \+ in the case the booleans are coherent *)

Lemma valid_joinE b1 b2 : 
        valid ([mx b1] \+ [mx b2]) -> 
        [mx b1] \+ [mx b2] = [mx b1 || b2].
Proof. by case: b1; case: b2. Qed.

Lemma validMx b1 b2 : valid ([mx b1] \+ [mx b2]) = b1 ==> ~~ b2.
Proof. by case: b1; case: b2. Qed.

(* Now the mutex resource algebra has to be tied to a pointer *)
(* in the heap that serves as the lock. *)

Section Shareable.
Variable x : ptr.

Definition shareable_axiom h := 
  exists v : lockb, exists m,
    [/\ h = x :-> v :+ m, def h &
        if v is Locked then m = empty else true].

(* parameter lock is the handle to the resource *)
(* METAVARIABLE lk, except in partitions, when we use sh, for shared *)
Structure shareable := 
  Shareable {
    (* the full heap, including the handle *)
    the_heap_of :> heap;
    (* the layout in the memory, and restrictions on ownerhip *)
    _ : shareable_axiom the_heap_of}.

Lemma shareablE (r1 r2 : shareable) : 
        r1 = r2 :> heap <-> r1 = r2.
Proof.
split=>[|-> //]; case: r1=>[h1 R1]; case: r2=>[h2 R2] /= T.
move: T R1 R2=>->=>R1 R2.
by congr Shareable; apply: proof_irrelevance.
Qed.

Definition payload (r : shareable) := free x r.

Lemma lockdT (r : shareable) : Dyn.typ (look x r) = lockb.
Proof. by case: r=>h [/= v][m][->] D _; rewrite lookPtUn. Qed.
  
Definition lockd (r : shareable) := locked
   match look x r as d return (Dyn.typ d = lockb -> _) with
     dyn A v => fun pf => coerce (T:=id) v pf == Locked
   end (lockdT r).

Lemma shrD (r : shareable) : def r.
Proof. by case: r=>h [/= v][m][->]. Qed.

Lemma shr_eta (r : shareable) : 
        r = x :-> (if lockd r then Locked
                   else Unlocked) :+ payload r :> heap.
Proof.
rewrite /lockd -ssreflect.lock /payload; move: (lockdT r).
case: r=>h [/= v][m][->] D O.
rewrite lookPtUn //; move=>pf; rewrite {pf}eqc freePtUn //.
by case: v D O.
Qed.

Lemma shrE (r1 r2 : shareable) : 
        [/\ lockd r1 = lockd r2 &
            payload r1 = payload r2] <-> r1 = r2.
Proof.
split; last by move=>->.
by case=>L P; apply/shareablE; rewrite 2!shr_eta L P.
Qed.

Lemma shr_null (r : shareable) : x != null. 
Proof. by move: (shrD r); rewrite shr_eta; apply: defPt_null. Qed.

Lemma shr_def r : def (payload r).
Proof. by move: (shrD r); rewrite shr_eta; apply: defPt_def. Qed.

Lemma shr_dom r : x \notin dom (payload r).
Proof. by move: (shrD r); rewrite shr_eta; apply: defPt_dom. Qed.

Lemma shr_ldom (r : shareable) : x \in dom r.
Proof. by rewrite shr_eta domPtUn !inE /= -shr_eta shrD eq_refl. Qed.

Lemma shrK (r1 r2 : shareable) h1 h2 : 
        def (r1 :+ h1) ->
        r1 :+ h1 = r2 :+ h2 -> 
        lockd r1 = lockd r2.
Proof.
move=>D E; rewrite !shr_eta -!unA in D E.
(* XXX probably can use look injectivity instead of cancel *)
by case/(cancel D): E; case: (lockd r1); case: (lockd r2).
Qed.

Lemma shr_setP (r : shareable) (v : lockb) h : 
        shareable_axiom 
        (if (v == Unlocked) && (def (x :-> v :+ h)) 
         then x :-> v :+ h else x :-> v).
Proof.
exists v, (if (v == Unlocked) && (def (x :-> v :+ h)) then h else empty).
case: v; first by rewrite unh0 /= defPt shr_null.
case: ifP=>/= E; first by rewrite E.
by rewrite unh0 defPt shr_null.
Qed.

Definition set_shr r v h := locked (Shareable (shr_setP r v h)).

Lemma lockdS r v h : 
        lockd (set_shr r v h) = 
        def (set_shr r v h) && (v == Locked).
Proof.
rewrite /set_shr /lockd -!ssreflect.lock /=; move: (lockdT _)=>/=.
case: v=>/=.
- rewrite lookU eq_refl /= (shr_null r).
  by move=>pf; rewrite eqc defPt shr_null.
case: ifP=>D.
- by rewrite lookPtUn //; move=>pf; rewrite eqc D.
rewrite lookU eq_refl (shr_null r) /=.
by move=>pf; rewrite eqc defPt shr_null.
Qed.

Lemma heapS r v h : 
        set_shr r v h = 
        (if [&& def h, x \notin dom h & v == Unlocked] 
         then x :-> v :+ h else x :-> v) :> heap.
Proof.
rewrite /set_shr -ssreflect.lock /=; case: ifP.
- by case/andP=>-> D; rewrite (defPt_dom D) (defPt_def D).
by rewrite defPtUn (shr_null r) /= andbC -andbA=>->.
Qed.

Lemma payloadS r v h : 
        payload (set_shr r v h) = 
        if [&& def h, x \notin dom h & v == Unlocked]
        then h else empty.
Proof.
rewrite /set_shr /payload -ssreflect.lock /=; case: ifP.
- by case/andP=>-> D; rewrite freePtUn // (defPt_dom D) (defPt_def D).
rewrite defPtUn (shr_null r) /= andbC -andbA=>->.
by rewrite freeU eq_refl (negbTE (shr_null r)) free0.
Qed.

Lemma payloadE r : lockd r -> payload r = empty.
Proof. 
rewrite /lockd /payload -ssreflect.lock; move: (lockdT _)=>/=.
case: r=>h [/= v][m][->] D O.
rewrite lookPtUn //; move=>pf; rewrite {pf}eqc freePtUn //.
by case: v D O.
Qed.

Definition receive r h := set_shr r Unlocked h.
Definition grant r := set_shr r Locked empty.

Lemma lockdR r h : lockd (receive r h) = false.
Proof. by rewrite lockdS shrD. Qed.

Lemma lockdG r : lockd (grant r).
Proof. by rewrite lockdS shrD. Qed.

Lemma payloadR (r : shareable) h : 
        def (r :+ h) -> payload (receive r h) = h.
Proof.
rewrite shr_eta unAC; move/defUnl; rewrite defPtUn; case/and3P.
by rewrite payloadS; move=>_ ->->.
Qed.

Lemma payloadG r : payload (grant r) = empty.
Proof. by apply: payloadE; apply: lockdG. Qed.

Lemma receiveE (r : shareable) h : 
        def (r :+ h) -> receive r h = x :-> Unlocked :+ h :> heap.
Proof. by move=>D; rewrite shr_eta lockdR payloadR. Qed.

Lemma grantE r : grant r = x :-> Locked :> heap.
Proof. by rewrite shr_eta lockdG payloadG unh0. Qed.

(* sometimes we need to inspect the lock value in the global heap with
an unknown partitioning, not just in the lheap part of the
partition. in that case, we don't have a decidable operation for
looking up the lock value, and we need a more general predicate. *)

Lemma shr_stores (r : shareable) : 
        x \In stores r (if lockd r then Locked else Unlocked).
Proof.
move: (shrD r); rewrite shr_eta => D.
by rewrite InE /= lookPtUn // domPtUn !inE /= D eq_refl.
Qed.

End Shareable.

(* Now we need a structure for combining heaps with ghost state, all
rolled into one. For lack of a better name, I will call these
process-private heaps, or just private heaps, in contrast to
shareables. They will consist of a heap, with some information about
the "contribution" to the heap by the process which owns it. *)

(* METAVARS: lp, ep, for local private heap, env private heap. When *)
(* generic, just use lp. *)

Section ProcessPrivateHeaps.
Variables U : pcm.

Structure procHeap := ProcHeap {mx : mutex; gh : U; hp :> heap}.

Lemma mghE (x1 x2 : procHeap) : 
       [/\ mx x1 = mx x2, gh x1 = gh x2 & x1 = x2 :> heap] <->
       x1 = x2.
Proof. by split=>[|-> //]; move: x1 x2=>[m1 g1 h1][m2 g2 h2][/= ->->->]. Qed.

(* as process heaps are just a combination of three different *)
(* resources, they form a pcm. *)

Definition phvalid := 
  fun x => [&& valid (mx x), valid (gh x) & valid (hp x)].

Definition phjoin := 
  fun x1 x2 => 
    ProcHeap (mx x1 \+ mx x2) (gh x1 \+ gh x2) (hp x1 \+ hp x2).

Definition phunit := ProcHeap Unit Unit Unit.

Local Notation "x \+ y" := (phjoin x y) (at level 43, left associativity).

Lemma phjoinC x y : x \+ y = y \+ x.
Proof. 
by rewrite /phjoin !(joinC (mx x)) !(joinC (gh x)) !(joinC (hp x)).
Qed.

Lemma phjoinA x y z : x \+ (y \+ z) = x \+ y \+ z.
Proof. by rewrite /phjoin !joinA. Qed.

(*
Lemma phjoinKr x y z : phvalid (x \+ y) -> x \+ y = x \+ z -> y = z.
Proof.
case/and3P=>V1 V2 V3 [E1 E2 E3]; apply/mghE.
by rewrite (joinKr V1 E1) (joinKr V2 E2) (joinKr V3 E3).
Qed.
*)

Lemma phvalidL x y : phvalid (x \+ y) -> phvalid x.
Proof.
case/and3P=>V1 V2 V3; rewrite /phvalid /=.
by rewrite (validL V1) (validL V2) (validL V3).
Qed.

Lemma phunitL x : phunit \+ x = x.
Proof. by rewrite /phjoin !unitL; apply/mghE. Qed.

Definition procHeapPCMMixin := PCMMixin phjoinC phjoinA phunitL phvalidL.
Canonical procHeapPCM := Eval hnf in PCM procHeap procHeapPCMMixin.

End ProcessPrivateHeaps.

(* projections and constructors *)

Notation mg g := (mx g, gh g).

Lemma mg_join U (lp1 lp2 : procHeap U) : 
        mg (lp1 \+ lp2) = mg lp1 \+ mg lp2.
Proof. by []. Qed.

Notation "[Mgh mx , gh , hp ]" := (ProcHeap mx gh hp)
  (at level 0, format "[Mgh  mx ,  gh ,  hp ]").

Notation "[mgh b , gh , hp ]" := ([Mgh [mx b], gh, hp]) 
  (at level 0, format "[mgh  b ,  gh ,  hp ]").

Lemma mgh_join (U : pcm) mx1 mx2 (gh1 gh2 : U) h1 h2 : 
        [Mgh mx1, gh1, h1] \+ [Mgh mx2, gh2, h2] = 
        [Mgh mx1 \+ mx2, gh1 \+ gh2, h1 \+ h2].
Proof. by []. Qed.

(* sometimes we need to expand in order for the simplification to *)
(* take place *)

Lemma mgh_eta (U : pcm) (lp : procHeap U) :
        lp = [Mgh mx lp, gh lp, lp].
Proof. by case: lp. Qed.

(* often we want to coerce a heap into a private heap, *)
(* while taking unit ghosts *)

Notation "[mg0 h ]" := ([Mgh Unit, Unit, h])
  (at level 0, format "[mg0  h ]").

(* Now we need to combine shareables and process-private heaps into one
big global state. *)

(*
Definition functional (U : pcm) (Inv : U -> heap -> Prop) :=
  forall h h1 h2 x1 x2 , 
    subheap h1 h -> valid x1 -> Inv x1 h1 -> 
    subheap h2 h -> valid x2 -> Inv x2 h2 -> 
    x1 = x2 /\ h1 = h2.
*)

Definition precise (Inv : heap -> Prop) :=
  forall h h1 h2,
    subheap h1 h -> Inv h1 ->
    subheap h2 h -> Inv h2 ->
    h1 = h2.

Definition functional_pred (A : Type) (P : Pred A) :=
  forall (x y : A), x \In P -> y \In P -> x = y.

(* Now a structure that packages a pcm, a pointer, and the invariant *)
(* It's not strictly needed right here, but without it, I ended up *)
(* constantly instantiating the lemmas from other files with the various *)
(* subsets of U, x, Inv, and func, instead of once-and-for-all instantiating *)
(* the few main concepts with the whole package, and then let the *)
(* inference figure out the rest *)

Structure resource := 
  Resource {univ_of : pcm;
            lock_of : ptr;

            (* module's heap invariant (ghost-independent) *)
            mod_inv_of : univ_of -> heap -> Prop;
            mod_inv_prec : forall g, precise (mod_inv_of g)}.

(* METAVAR s *)
Section StatePartition.
Variable R : resource.

Notation U := (univ_of R).
Notation x := (lock_of R).
Notation shareable := (shareable x).
Notation procHeap := (procHeap U).
Notation mod_inv := (@mod_inv_of R).
Notation mod_inv_prec := (@mod_inv_prec R).
(* Notation repn := (@repn_of R). *)
(* Notation next := (@next_of R). *)
(* Notation cli_inv := (@cli_inv_of R). *)
(* Notation Inv := (@inv_of R). *)
(* Notation prec := (@mod_inv_prec R). *)
(* Notation func := (@func_of R). *)

(* XXX alternative arrangement of conjuncts? view between Inv/Inv'? *)
Definition Inv := mod_inv.
  (* fun g h => mod_inv h /\ exists rp, rp \In repn h /\ cli_inv rp g. *)

(*
Definition Inv' gE gL gL' h :=
  mod_inv h /\
  exists rp,
    [/\ rp \In repn h, gL' \In next rp gL gE & cli_inv rp (gL' \+ gE)].

Lemma inv_assoc g gL gL' gE h :
        Inv' (g \+ gE) gL gL' h -> Inv' gE (gL \+ g) (gL' \+ g) h.
Proof.
by case=>MI [rp] [RI]; move/next_assoc=>N; rewrite joinA; split=>//; exists rp.
Qed.
*)

(*
Lemma funcE sh1 sh2 ep1 ep2 h1 h2 (x1 x2 : U) : 
        def (sh1 :+ ep1) -> sh1 :+ ep1 = sh2 :+ ep2 -> 
        subheap h1 sh1 -> valid x1 -> Inv x1 h1 -> 
        subheap h2 sh2 -> valid x2 -> Inv x2 h2 -> 
        x1 = x2 /\ h1 = h2.
Proof.
move=>D E Sh1 V1 [MI1] [rp1] [R1] CI1 Sh2 V2 [MI2] [rp2] [R2] CI2.
have H1: subheap h1 (sh1 :+ ep1).
- by apply: subheap_trans Sh1 _; apply: subheapUnl.
have H2: subheap h2 (sh1 :+ ep1).
- by apply: subheap_trans Sh2 _; rewrite E in D *; apply: subheapUnl.
rewrite -{h2 E Sh1 MI1 H1 Sh2 MI2 H2}(mod_inv_prec H1 MI1 H2 MI2) in R2 *.
rewrite -{rp2 R1 R2}(repn_func R1 R2) in CI2.


by apply: func H1 V1 I1 H2 V2 I2.
Qed.
*)

(*
Lemma precE sh1 sh2 ep1 ep2 h1 h2 : 
        def (sh1 :+ ep1) -> sh1 :+ ep1 = sh2 :+ ep2 -> 
        subheap h1 sh1 -> mod_inv h1 -> 
        subheap h2 sh2 -> mod_inv h2 -> 
        h1 = h2.
Proof.
move=>D E Sh1 I1 Sh2 I2.
have H1: subheap h1 (sh1 :+ ep1).
- by apply: subheap_trans Sh1 _; apply: subheapUnl.
have H2: subheap h2 (sh1 :+ ep1).
- by apply: subheap_trans Sh2 _; rewrite E in D *; apply: subheapUnl.
by apply: mod_inv_prec H1 I1 H2 I2.
Qed.
*)

(*Lemma funcE sh1 sh2 ep1 ep2 h1 h2 (x1 x2 : U) : 
        def (sh1 :+ ep1) -> sh1 :+ ep1 = sh2 :+ ep2 -> 
        subheap h1 sh1 -> valid x1 -> Inv x1 h1 -> 
        subheap h2 sh2 -> valid x2 -> Inv x2 h2 -> 
        x1 = x2 /\ h1 = h2.
Proof.
move=>D E Sh1 V1 I1 Sh2 V2 I2.
have H1: subheap h1 (sh1 :+ ep1).
- by apply: subheap_trans Sh1 _; apply: subheapUnl.
have H2: subheap h2 (sh1 :+ ep1).
- by apply: subheap_trans Sh2 _; rewrite E in D *; apply: subheapUnl.
by apply: func H1 V1 I1 H2 V2 I2.
Qed.*)

Structure state := 
  StatePartition {shared : shareable; local : procHeap; env : procHeap}.

(* constructors *)

Notation "[ sh , lp , ep ]" :=
  (StatePartition sh lp ep)
  (at level 0, format "[ sh ,  lp ,  ep ]").

Lemma st_eta s : s = [shared s, local s, env s].
Proof. by case: s. Qed.

Lemma st_feta A (f : state -> A) : 
        f = fun s => f [shared s, local s, env s].
Proof. by apply: fext; case. Qed.

Definition heap_of := 
  [fun s : state => shared s :+ local s :+ env s].

Lemma stateK (sh1 sh2 : shareable) hL1 hL2 hE1 hE2 :
        def (sh1 :+ hL1 :+ hE1) -> 
        sh1 :+ hL1 :+ hE1 = sh2 :+ hL2 :+ hE2 -> 
        lockd sh1 = lockd sh2.
Proof. by rewrite -!unA; apply: shrK. Qed.

(* common rewrites when transfering parts of the heap *)

Lemma shep (sh : shareable) hL hE :
        upd (sh :+ hL :+ hE) x (dyn Locked) = 
        grant sh :+ hL :+ (hE :+ payload sh).
Proof. by rewrite -!unA updUnl shr_ldom grantE shr_eta updPtUn unAC !unA. Qed.

Lemma shlp (sh : shareable) hL hE :
        upd (sh :+ hL :+ hE) x (dyn Locked) = 
        grant sh :+ (hL :+ payload sh) :+ hE.
Proof. by rewrite -!(unAC _ hE) shep. Qed.

Lemma epsh (sh : shareable) hL hE inv : 
        [/\ def (sh :+ hL :+ hE), lockd sh & subheap inv hE] ->
        upd (sh :+ hL :+ hE) x (dyn Unlocked) = 
        receive sh inv :+ hL :+ (hE :- inv).
Proof.
case=>D L Sh; rewrite -!unA updUnl shr_ldom receiveE; last first.
- by case/subheapX: Sh D=>-> _; defcheck.
rewrite shr_eta updPtUn payloadE // unh0.
by rewrite unAC !unA -(unA _ _ inv) (subheapK Sh).
Qed.

Lemma lpsh (sh : shareable) hL hE inv : 
        [/\ def (sh :+ hL :+ hE), lockd sh & subheap inv hL] ->
        upd (sh :+ hL :+ hE) x (dyn Unlocked) = 
        receive sh inv :+ (hL :- inv) :+ hE.
Proof. by rewrite -!(unAC _ hE); apply: epsh. Qed.

(* A state partition is essentially a combination of three different
resources: mutex, arbitrary ghost, and a heap, and that for two
processes.  But it is not a simple cartesian product of these three
resource algebras. These are not independent, and we need an
additional condition to select only those triples that are coherent,
that is, respect the intended invariants. *)

(* XXX this could be made more uniform as follows:
   1. define a sub-pcm of heap * mutex * ghost
      with the triple of (lcl, shr, env)
      where lcl = (hL, mL, gL)
            shr = (hS, unit, unit)
            lcl = (hE, mE, gE)
   2. define validity of the sub-pcm as
   2a. valid (lcl s \+ shr s \+ env s)
   2b. owns (mx (lcl\+shr\+env)) ==> hS = lk :-> Unlocked
   2c. ~owns (mx (lcl\+shr\+env)) ==> hS = lk :-> Locked :+ hInv /\ Inv hInv
*)
Definition coherent := 
  [Pred s | [/\ def (heap_of s),
                valid (mg (local s) \+ mg (env s)),
                owns (mx (local s)) || owns (mx (env s)) ==> lockd (shared s) 
              & ~~ lockd (shared s) -> 
                Inv (gh (local s) \+ gh (env s)) (payload (shared s))]].

Lemma cohD s : s \In coherent -> def (heap_of s).
Proof. by case. Qed.

Lemma cohC sh lp ep : 
       [sh, lp, ep] \In coherent <-> [sh, ep, lp] \In coherent.
Proof. 
by split; case=>/=; rewrite unAC (joinC (mg _)) orbC (joinC (gh _)).
Qed.

Lemma cohV sh lp ep : 
        [sh, lp, ep] \In coherent -> valid (mg lp \+ mg ep).
Proof. by case. Qed.

Lemma cohVgh sh lp ep : 
        [sh, lp, ep] \In coherent -> valid (gh lp \+ gh ep).
Proof. by move/cohV/andP/proj2. Qed.

(* combination of cohV and a subset of cohD *)
Lemma cohVld (sh : shareable) lp ep :  
        [sh, lp, ep] \In coherent -> valid (lp \+ ep).
Proof. 
case=>/= D; case/andP=>/= V1 V2 _ _.
apply/and3P; rewrite V1 V2 /valid /=.
by move: D; rewrite -unA; move/defUnr.
Qed.

(* excluded invalid *)

Lemma coh_exL sh lp ep : 
        [sh, lp, ep] \In coherent -> owns (mx lp) || ownsno (mx lp).
Proof. by case/cohV/andP; case: (mx lp). Qed.

Lemma coh_exE sh lp ep : 
        [sh, lp, ep] \In coherent -> owns (mx ep) || ownsno (mx ep).
Proof. by move/cohC/coh_exL. Qed.

(* ownerships are exclusive *)

Lemma coh_ownsL sh lp ep : 
         [sh, lp, ep] \In coherent -> owns (mx lp) ==> ownsno (mx ep).
Proof. by move/cohV; case: (mx lp); case: (mx ep). Qed.

Lemma coh_ownsE sh lp ep : 
         [sh, lp, ep] \In coherent -> owns (mx ep) ==> ownsno (mx lp).
Proof. by move/cohC; move/coh_ownsL. Qed.

(* if mutex owned, then the lock is lockd *)

Lemma coh_owns_lkd sh lp ep : 
       [sh, lp, ep] \In coherent -> 
       owns (mx lp) || owns (mx ep) ==> lockd sh.
Proof. by case. Qed.

(* the "contrapositive"; excludes the invalid cases *)

Lemma coh_lkd_owns sh lp ep :
        [sh, lp, ep] \In coherent -> 
        ~~ lockd sh ==> ownsno (mx lp) && ownsno (mx ep).
Proof.
move=>C; move: (coh_owns_lkd C) (coh_exL C) (coh_exE C).
by case: (mx lp); case: (mx ep); case: (lockd sh).
Qed.

(* Do we have framing wrt mL and gL? *)

Lemma coh_frame (sh : shareable) (lp ep : procHeap) h : 
        def (sh :+ lp :+ ep :+ h) ->
        [sh, lp, ep] \In coherent ->
        [sh, lp \+ [mg0 h], ep] \In coherent.
Proof. 
by move=>D; case=>/= _ V J I; rewrite InE /= unA unAC D !unitR V J.
Qed.

Lemma frame_coh (sh : shareable) (lp ep : procHeap) h : 
        [sh, lp \+ [mg0 h], ep] \In coherent ->
        [sh, lp, ep] \In coherent.
Proof. 
by case=>/=; rewrite !unitR=>D V O J; split=>//=; move: D; defcheck.
Qed.

Lemma coh_sub (sh : shareable) (lp1 lp2 ep1 ep2 : procHeap) : 
        [/\ def (sh :+ lp2 :+ ep2), 
            mg lp1 = mg lp2 & mg ep1 = mg ep2] ->
        [sh, lp1, ep1] \In coherent ->
        [sh, lp2, ep2] \In coherent.
Proof. by rewrite !InE /=; case=>D [->->][->->][]. Qed.

Lemma cohRlp (sh : shareable) gL hL (ep : procHeap) inv : 
        [/\ def (sh :+ hL :+ ep), valid (gL \+ gh ep), ownsno (mx ep), 
            lockd sh, subheap inv hL & Inv (gL \+ gh ep) inv] -> 
        [receive sh inv, [mgh false, gL, hL :- inv], ep] \In coherent.
Proof.
case=>D V O L Sh I; rewrite InE /= /valid lockdR (eqP O) /=.
rewrite -lpsh // defU (shr_null sh) D payloadR //.
by case/subheapX: Sh D=>-> _; defcheck.
Qed.

Lemma cohRep (sh : shareable) (lp : procHeap) gE hE inv : 
        [/\ def (sh :+ lp :+ hE), valid (gh lp \+ gE), ownsno (mx lp), 
            lockd sh, subheap inv hE & Inv (gh lp \+ gE) inv] -> 
        [receive sh inv, lp, [mgh false, gE, hE :- inv]] \In coherent.
Proof. by rewrite unAC joinC cohC; apply: cohRlp. Qed.

Lemma cohGlp (sh : shareable) gL hL (ep : procHeap) :
        [/\ def (sh :+ hL :+ ep), valid (gL \+ gh ep) & ownsno (mx ep)] ->
        [grant sh, [mgh true, gL, hL :+ payload sh], ep] \In coherent.
Proof.
case=>D V O; rewrite InE /= /valid lockdG (eqP O) /=.
by rewrite -shlp defU (shr_null sh).
Qed.

Lemma cohGep (sh : shareable) (lp : procHeap) gE hE :
        [/\ def (sh :+ lp :+ hE), valid (gh lp \+ gE) & ownsno (mx lp)] ->
        [grant sh, lp, [mgh true, gE, hE :+ payload sh]] \In coherent.
Proof. by rewrite unAC joinC cohC; apply: cohGlp. Qed.

Lemma coh_lockd (sh1 sh2 : shareable) lp1 lp2 ep1 ep2 :
        [sh1, lp1, ep1] \In coherent ->
        sh1 :+ lp1 :+ ep1 = sh2 :+ lp2 :+ ep2 ->
        lockd sh1 = lockd sh2.
Proof. by move/cohD=>D; apply: stateK. Qed.

(*Lemma cohKlp sh1 sh2 lp ep1 ep2 : 
        [sh1, lp, ep1] \In coherent ->
        [sh2, lp, ep2] \In coherent ->
        sh1 :+ lp :+ ep1 = sh2 :+ lp :+ ep2 -> 
        [/\ sh1 = sh2 & ep1 = ep2 :> heap].
Proof.
case=>/= D V1 _ J1 [/= _ V2 _ J2] E.
case/andP: V1=>/= _ V1; case/andP: V2=>/= _ V2.
move: (stateK D E)=>E3; rewrite -!(unC lp) -!unA in D E.
move/(unhKl D): E=>E; move/defUnr: D=>D.
suff P: payload sh1 = payload sh2.
- rewrite (shr_eta sh1) (shr_eta sh2) E3 P -!unA in D E.
  case/(cancel D): E=>{D} _ D; move/(unhKl D)=>->; split=>//.
  by apply/shrE.
case L2: (lockd sh2) E3=>L1; first by rewrite !payloadE.
suff S: forall (sh : shareable), subheap (payload sh) sh.
have E' : gh ep2 = gh ep1 by admit.(**)
rewrite E' in J2.
by apply: (precE D E (S _) (J1 (negbT L1)) (S _) (J2 (negbT L2))).
(* - by case: (funcE D E (S _) V1 (J1 (negbT L1)) (S _) V2 (J2 (negbT L2))). *)
move=>sh; rewrite (shr_eta sh).
by apply: subheapUnr; rewrite -shr_eta shrD.
Qed.

Lemma cohKep sh1 sh2 lp1 lp2 ep : 
        [sh1, lp1, ep] \In coherent ->
        [sh2, lp2, ep] \In coherent ->
        sh1 :+ lp1 :+ ep = sh2 :+ lp2 :+ ep -> 
        [/\ sh1 = sh2 & lp1 = lp2 :> heap].
Proof.
rewrite !(cohC _ _ ep) -!(unAC _ ep) sym => C1 C2.
by case/(cohKlp C2 C1)=>->->.
Qed.*)

(* if we have a third process with its own local state mp we can make
mp be an environment for the process running to the left of our
process, iff we can make it environment for process running to the
right *)

Lemma coh_shuffle sh lp ep mp :
        [sh, lp \+ mp, ep] \In coherent <-> 
        [sh, lp, mp \+ ep] \In coherent.
Proof.
suff {lp ep mp} H: forall lp ep mp, 
  [sh, lp \+ mp, ep] \In coherent -> [sh, lp, mp \+ ep] \In coherent.
- split; first by apply: H.
  by rewrite joinC; move/cohC/H; rewrite joinC; move/cohC.
move=>lp ep mp; case=>/= D; case/andP=>/= V1 V2 H I.
rewrite InE /= !heap_join !unA in D *.
rewrite D /valid /= !joinA V1 V2; split=>//; move: H.
rewrite own_joinE; last by apply: validL V1.
rewrite own_joinE; first by rewrite orbA.
by rewrite -joinA in V1; apply: validR V1.
Qed.

End StatePartition.

(* restate the notation *)

Notation "[ sh , lp , ep ]" :=
  (StatePartition sh lp ep)
  (at level 0, format "[ sh ,  lp ,  ep ]").

Implicit Arguments coherent [R].
Implicit Arguments heap_of [R].
